"""Tests for grizzly_cli.utils."""

from __future__ import annotations

from argparse import Namespace
from contextlib import ExitStack
from importlib import reload
from json.decoder import JSONDecodeError
from pathlib import Path
from tempfile import gettempdir
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Union
from unittest.mock import mock_open
from unittest.mock import patch as unittest_patch

import pytest
from grizzly_cli.utils import (
    ask_yes_no,
    distribution_of_users_per_scenario,
    find_metadata_notices,
    find_variable_names_in_questions,
    get_default_mtu,
    get_dependency_versions,
    get_distributed_system,
    list_images,
    parse_feature_file,
    requirements,
    run_command,
    setup_logging,
)

from test_cli.helpers import create_scenario, cwd, rm_rf

if TYPE_CHECKING:  # pragma: no cover
    from _pytest.capture import CaptureFixture
    from _pytest.tmpdir import TempPathFactory
    from pytest_mock import MockerFixture
    from requests_mock import Mocker as RequestsMocker


pytest_plugins = ['requests_mock']


def test_parse_feature_file(tmp_path_factory: TempPathFactory) -> None:
    test_context = tmp_path_factory.mktemp('test_context')
    test_context_root = str(test_context)
    feature_file = test_context / 'test.feature'
    feature_file.touch()
    feature_file.write_text(
        dedent("""
    Feature: test feature
        Background:
            Given a common test step
            When executed in every scenario
        Scenario: scenario-1
            Given a test step
            And another test step
        Scenario: scenario-2
            Given a second test step
            Then execute it
            When done, just stop
    """)
    )

    try:
        with cwd(test_context):
            import grizzly_cli

            reload(grizzly_cli)
            reload(grizzly_cli.utils)

            assert len(grizzly_cli.SCENARIOS) == 0

            parse_feature_file('test.feature')

            import grizzly_cli

            cached_scenarios = grizzly_cli.SCENARIOS.copy()
            assert len(grizzly_cli.SCENARIOS) == 2
            assert next(iter(grizzly_cli.SCENARIOS)).name == 'scenario-1'
            assert len(next(iter(grizzly_cli.SCENARIOS)).steps) == 2
            assert len(next(iter(grizzly_cli.SCENARIOS)).background_steps) == 2
            assert list(grizzly_cli.SCENARIOS)[1].name == 'scenario-2'
            assert len(list(grizzly_cli.SCENARIOS)[1].steps) == 3
            assert len(list(grizzly_cli.SCENARIOS)[1].background_steps) == 2

            parse_feature_file('test.feature')

            import grizzly_cli

            assert cached_scenarios == grizzly_cli.SCENARIOS
    finally:
        rm_rf(test_context_root)


def test_list_images(mocker: MockerFixture) -> None:
    check_output = mocker.patch(
        'grizzly_cli.utils.subprocess.check_output',
        side_effect=[
            (
                b'{"name": "mcr.microsoft.com/vscode/devcontainers/python", "tag": "0-3.10", "size": "1.16GB", "created": "2021-12-02 23:46:55 +0100 CET", "id": "a05f8cc8454b"}\n'
                b'{"name": "mcr.microsoft.com/vscode/devcontainers/python", "tag": "0-3.10-bullseye", "size": "1.16GB", "created": "2021-12-02 23:46:55 +0100 CET", "id": "a05f8cc8454b"}\n'  # noqa: E501
                b'{"name": "mcr.microsoft.com/vscode/devcontainers/python", "tag": "0-3.9", "size": "1.23GB", "created": "2021-12-02 23:27:50 +0100 CET", "id": "bfbce224d490"}\n'
                b'{"name": "mcr.microsoft.com/vscode/devcontainers/python", "tag": "0-3.8", "size": "1.23GB", "created": "2021-12-02 23:10:12 +0100 CET", "id": "8a04d9e5df14"}\n'
                b'{"name": "mcr.microsoft.com/vscode/devcontainers/base", "tag": "0-focal", "size": "343MB", "created": "2021-12-02 22:44:23 +0100 CET", "id": "0cc1cbb6d08d"}\n'
                b'{"name": "mcr.microsoft.com/vscode/devcontainers/python", "tag": "0-3.6", "size": "1.22GB", "created": "2021-12-02 22:17:47 +0100 CET", "id": "cc5abbf52b04"}\n'
            )
        ],
    )

    arguments = Namespace(container_system='capsulegirl')

    images = list_images(arguments)

    assert check_output.call_count == 1
    args, _ = check_output.call_args_list[-1]
    assert args[0] == [
        'capsulegirl',
        'image',
        'ls',
        '--format',
        '{"name": "{{.Repository}}", "tag": "{{.Tag}}", "size": "{{.Size}}", "created": "{{.CreatedAt}}", "id": "{{.ID}}"}',
    ]

    assert len(images.keys()) == 2
    assert sorted(images.get('mcr.microsoft.com/vscode/devcontainers/python', {}).keys()) == sorted(
        [
            '0-3.10',
            '0-3.10-bullseye',
            '0-3.9',
            '0-3.8',
            '0-3.6',
        ]
    )
    assert sorted(images.get('mcr.microsoft.com/vscode/devcontainers/base', {}).keys()) == sorted(
        [
            '0-focal',
        ]
    )


def test_get_default_mtu(mocker: MockerFixture) -> None:
    check_output = mocker.patch(
        'grizzly_cli.utils.subprocess.check_output',
        side_effect=[
            JSONDecodeError,
            (
                b'{"com.docker.network.bridge.default_bridge":"true","com.docker.network.bridge.enable_icc":"true",'
                b'"com.docker.network.bridge.enable_ip_masquerade":"true","com.docker.network.bridge.host_binding_ipv4":"0.0.0.0",'
                b'"com.docker.network.bridge.name":"docker0","com.docker.network.driver.mtu":"1500"}\n'
            ),
            (
                b'{"com.docker.network.bridge.default_bridge":"true","com.docker.network.bridge.enable_icc":"true",'
                b'"com.docker.network.bridge.enable_ip_masquerade":"true","com.docker.network.bridge.host_binding_ipv4":"0.0.0.0",'
                b'"com.docker.network.bridge.name":"docker0","com.docker.network.driver.mtu":"1440"}\n'
            ),
        ],
    )

    arguments = Namespace(container_system='capsulegirl')

    assert get_default_mtu(arguments) is None  # JSONDecodeError

    assert check_output.call_count == 1
    args, _ = check_output.call_args_list[-1]
    assert args[0] == [
        'capsulegirl',
        'network',
        'inspect',
        'bridge',
        '--format',
        '{{ json .Options }}',
    ]

    assert get_default_mtu(arguments) == '1500'
    assert get_default_mtu(arguments) == '1440'

    assert check_output.call_count == 3


def test_run_command(capsys: CaptureFixture, mocker: MockerFixture) -> None:
    setup_logging()

    terminate = mocker.patch('grizzly_cli.utils.subprocess.Popen.terminate', autospec=True)
    wait = mocker.patch('grizzly_cli.utils.subprocess.Popen.wait', autospec=True)

    def popen___init___no_stdout(*args: Any, **_kwargs: Any) -> None:
        args[0].returncode = 133
        args[0].stdout = None

    mocker.patch('grizzly_cli.utils.subprocess.Popen.__init__', popen___init___no_stdout)
    poll_mock = mocker.patch('grizzly_cli.utils.subprocess.Popen.poll', side_effect=[None])
    kill_mock = mocker.patch('grizzly_cli.utils.subprocess.Popen.kill', side_effect=[RuntimeError, None])

    assert run_command(['hello', 'world'], verbose=True).return_code == 133

    capture = capsys.readouterr()
    assert capture.out == ''
    assert capture.err == 'run_command: hello world\n'

    assert terminate.call_count == 1
    assert wait.call_count == 1
    assert poll_mock.call_count == 1
    assert kill_mock.call_count == 1

    def mock_command_output(output: list[str], returncode: int = 0) -> None:
        output_buffer: list[Union[bytes, int]] = [f'{line}\n'.encode() for line in output] + [0]

        def popen___init__(*args: Any, **_kwargs: Any) -> None:
            args[0].returncode = returncode

            class Stdout:
                def readline(self) -> Union[bytes, int]:
                    return output_buffer.pop(0)

            args[0].stdout = Stdout()

        mocker.patch('grizzly_cli.utils.subprocess.Popen.terminate', side_effect=[KeyboardInterrupt])
        mocker.patch('grizzly_cli.utils.subprocess.Popen.__init__', popen___init__)

    mock_command_output(
        [
            'first line',
            'second line',
        ]
    )
    poll_mock = mocker.patch('grizzly_cli.utils.subprocess.Popen.poll', side_effect=[None] * 3)

    result = run_command([], {})
    assert result.return_code == 0
    assert result.output is None
    assert result.abort_timestamp is None

    capture = capsys.readouterr()
    assert capture.out == ''
    assert capture.err == ('first line\nsecond line\n')

    assert wait.call_count == 2
    assert poll_mock.call_count == 3
    assert kill_mock.call_count == 2

    mock_command_output(
        [
            'hello world',
            'foo bar',
            'bar grizzly.returncode=1234 foo',
            'grizzly.returncode=4321',
            'world foo hello bar',
        ],
        4321,
    )
    poll_mock = mocker.patch('grizzly_cli.utils.subprocess.Popen.poll', side_effect=[None] * 6)

    result = run_command([], {}, silent=True)
    assert result.return_code == 4321
    assert result.output == [
        b'hello world\n',
        b'foo bar\n',
        b'bar grizzly.returncode=1234 foo\n',
        b'grizzly.returncode=4321\n',
        b'world foo hello bar\n',
    ]

    capture = capsys.readouterr()
    assert capture.err == ''
    assert capture.out == ''

    assert wait.call_count == 3
    assert poll_mock.call_count == 6
    assert kill_mock.call_count == 3


def test_get_distributed_system(capsys: CaptureFixture, mocker: MockerFixture) -> None:
    which = mocker.patch('grizzly_cli.utils.which')
    getstatusoutput = mocker.patch('grizzly_cli.utils.subprocess.getstatusoutput')

    # test 1
    which.side_effect = [None, None]
    getstatusoutput.return_value = (1, 'foobar')
    assert get_distributed_system() is None  # neither
    capture = capsys.readouterr()
    assert capture.out == 'neither "podman" nor "docker" found in PATH\n'
    assert which.call_count == 2
    getstatusoutput.assert_not_called()
    which.reset_mock()

    # test 2
    which.side_effect = [None, 'podman']
    getstatusoutput.return_value = (1, 'foobar')
    assert get_distributed_system() is None
    capture = capsys.readouterr()
    assert which.call_count == 2
    getstatusoutput.assert_called_once_with('podman compose version')
    assert capture.out == (
        '!! podman might not work due to buildah missing support for `RUN --mount=type=ssh`: https://github.com/containers/buildah/issues/2835\n'
        '"podman compose" not found in PATH\n'
    )
    which.reset_mock()
    getstatusoutput.reset_mock()

    # test 3
    which.side_effect = [None, 'podman']
    getstatusoutput.return_value = (0, 'foobar')
    assert get_distributed_system() == 'podman'
    capture = capsys.readouterr()
    assert which.call_count == 2
    getstatusoutput.assert_called_once_with('podman compose version')
    assert capture.out == ('!! podman might not work due to buildah missing support for `RUN --mount=type=ssh`: https://github.com/containers/buildah/issues/2835\n')
    which.reset_mock()
    getstatusoutput.reset_mock()

    # test 4
    which.side_effect = ['docker']
    getstatusoutput.return_value = (1, 'foobar')
    assert get_distributed_system() is None
    capture = capsys.readouterr()
    assert which.call_count == 1
    getstatusoutput.assert_called_once_with('docker compose version')
    assert capture.out == ('"docker compose" not found in PATH\n')
    which.reset_mock()
    getstatusoutput.reset_mock()

    # test 5
    which.side_effect = ['docker']
    getstatusoutput.return_value = (0, 'foobar')
    assert get_distributed_system() == 'docker'
    capture = capsys.readouterr()
    assert which.call_count == 1
    getstatusoutput.assert_called_once_with('docker compose version')
    assert capture.out == ''
    which.reset_mock()


def test_find_variable_names_in_questions(mocker: MockerFixture) -> None:
    mocker.patch('grizzly_cli.SCENARIOS', [])
    mocker.patch('grizzly_cli.utils.parse_feature_file', autospec=True)

    assert find_variable_names_in_questions('test.feature') == []

    mocker.patch(
        'grizzly_cli.SCENARIOS',
        [
            create_scenario(
                'scenario-1',
                [],
                [
                    'Given a user of type "RestApi" load testing "https://localhost"',
                    'And ask for value of variable test_variable_1',
                ],
            ),
        ],
    )

    with pytest.raises(ValueError, match='could not find variable name in "ask for value of variable test_variable_1'):
        find_variable_names_in_questions('test.feature')

    mocker.patch(
        'grizzly_cli.SCENARIOS',
        [
            create_scenario(
                'scenario-1',
                [],
                [
                    'Given a user of type "RestApi" load testing "https://localhost"',
                    'And ask for value of variable "test_variable_2"',
                    'And ask for value of variable "test_variable_1"',
                ],
            ),
            create_scenario(
                'scenario-2',
                [
                    'And ask for value of variable "bar"',
                ],
                [
                    'Given a user of type "MessageQueueUser" load testing "mqs://localhost"',
                    'And ask for value of variable "foo"',
                ],
            ),
        ],
    )
    variables = find_variable_names_in_questions('test.feature')
    assert len(variables) == 4
    assert variables == ['bar', 'foo', 'test_variable_1', 'test_variable_2']


def test_find_metadata_notices(tmp_path_factory: TempPathFactory) -> None:
    test_context = tmp_path_factory.mktemp('test_context')

    try:
        feature_file = test_context / 'test-1.feature'
        feature_file.write_text("""Feature: test -1
    Scenario: hello world
        Given a feature file with a rich set of expressions
""")
        assert find_metadata_notices(str(feature_file)) == []

        feature_file.write_text("""# grizzly-cli run --verbose
# grizzly-cli:notice have you created testdata?
Feature: test -1
    Scenario: hello world
        Given a feature file with a rich set of expressions
""")

        assert find_metadata_notices(str(feature_file)) == ['have you created testdata?']

        feature_file.write_text("""# grizzly-cli run --verbose
# grizzly-cli:notice have you created testdata?
Feature: test -1
    Scenario: hello world
        # grizzly-cli:notice is the event log cleared?
        Given a feature file with a rich set of expressions
""")

        assert find_metadata_notices(str(feature_file)) == ['have you created testdata?', 'is the event log cleared?']
    finally:
        rm_rf(test_context)


def test_distribution_of_users_per_scenario(capsys: CaptureFixture, mocker: MockerFixture) -> None:  # noqa: PLR0915
    setup_logging()

    arguments = Namespace(file='test.feature', yes=False)

    ask_yes_no = mocker.patch('grizzly_cli.utils.ask_yes_no', autospec=True)

    mocker.patch(
        'grizzly_cli.SCENARIOS',
        [
            create_scenario(
                'scenario-1',
                [],
                [
                    'Given a user of type "RestApi" load testing "https://localhost"',
                    'And ask for value of variable "test_variable_2"',
                    'And ask for value of variable "test_variable_1"',
                ],
            ),
        ],
    )

    with pytest.raises(ValueError, match='grizzly needs at least 1 users to run this feature'):
        distribution_of_users_per_scenario(arguments, {})

    mocker.patch(
        'grizzly_cli.SCENARIOS',
        [
            create_scenario(
                'scenario-1',
                [
                    'Given "10" users',
                ],
                [
                    'Given a user of type "RestApi" load testing "https://localhost"',
                    'And repeat for "1" iteration',
                    'And ask for value of variable "test_variable_2"',
                    'And ask for value of variable "test_variable_1"',
                ],
            ),
            create_scenario(
                'scenario-2',
                [
                    'And ask for value of variable "bar"',
                ],
                [
                    'Given a user of type "MessageQueueUser" load testing "mqs://localhost"',
                    'And repeat for "1" iteration',
                    'And ask for value of variable "foo"',
                ],
            ),
        ],
    )

    with pytest.raises(ValueError, match='scenario-1 will have 5 users to run 1 iterations, increase iterations or lower user count'):
        distribution_of_users_per_scenario(arguments, {})

    capsys.readouterr()

    mocker.patch(
        'grizzly_cli.SCENARIOS',
        [
            create_scenario(
                'scenario-1',
                [
                    'Given "2" users',
                ],
                [
                    'Given a user of type "RestApi" load testing "https://localhost"',
                    'And repeat for "1" iteration',
                    'And ask for value of variable "test_variable_2"',
                    'And ask for value of variable "test_variable_1"',
                ],
            ),
            create_scenario(
                'scenario-2',
                [
                    'And ask for value of variable "bar"',
                ],
                [
                    'Given a user of type "MessageQueueUser" load testing "mqs://localhost"',
                    'And repeat for "1" iteration',
                    'And ask for value of variable "foo"',
                ],
            ),
        ],
    )

    distribution_of_users_per_scenario(arguments, {})
    capture = capsys.readouterr()

    assert capture.out == ''
    expected_lines = [
        '\n',
        'feature file test.feature will execute in total 2 iterations divided on 2 scenarios\n',
        '\n',
        'each scenario will execute accordingly:\n',
        '\n',
        'ident   weight  #iter  #user  description\n',
        '------|-------|------|------|-------------|\n',
        '001          1      1      1  scenario-1 \n',
        '002          1      1      1  scenario-2 \n',
        '------|-------|------|------|-------------|\n',
        '\n',
    ]
    assert capture.err == ''.join(expected_lines)
    capsys.readouterr()
    assert ask_yes_no.call_count == 1
    args, _ = ask_yes_no.call_args_list[-1]
    assert args[0] == 'continue?'

    mocker.patch(
        'grizzly_cli.SCENARIOS',
        [
            create_scenario(
                'scenario-1',
                [],
                [],
            ),
        ],
    )

    with pytest.raises(ValueError, match='scenario "scenario-1" does not have any steps'):
        distribution_of_users_per_scenario(arguments, {})

    mocker.patch(
        'grizzly_cli.SCENARIOS',
        [
            create_scenario(
                'scenario-1',
                ['Given "1" users'],
                ['And repeat for "10" iterations'],
            ),
        ],
    )

    with pytest.raises(ValueError, match='scenario-1 does not have a user type'):
        distribution_of_users_per_scenario(arguments, {})

    mocker.patch(
        'grizzly_cli.SCENARIOS',
        [
            create_scenario(
                'scenario-1',
                [
                    'Given "{{ users }}" users',
                ],
                [
                    'Given a user of type "RestApi" with weight "{{ integer }}" load testing "https://localhost"',
                    'And repeat for "{{ integer * 0.10 }}" iterationsAnd ask for value of variable "test_variable_2"',
                    'And ask for value of variable "test_variable_1"',
                ],
            ),
            create_scenario(
                'scenario-2',
                [
                    'And ask for value of variable "bar"',
                ],
                [
                    'Given a user of type "MessageQueueUser" load testing "mqs://localhost"',
                    'And repeat for "{{ integer * 0.01 }}" iterations',
                    'And ask for value of variable "foo"',
                ],
            ),
        ],
    )

    import grizzly_cli.utils

    render = mocker.spy(grizzly_cli.utils.Template, 'render')  # type: ignore[attr-defined]

    distribution_of_users_per_scenario(
        arguments,
        {
            'TESTDATA_VARIABLE_users': '40',
            'TESTDATA_VARIABLE_boolean': 'True',
            'TESTDATA_VARIABLE_integer': '500',
            'TESTDATA_VARIABLE_float': '1.33',
            'TESTDATA_VARIABLE_string': 'foo bar',
            'TESTDATA_VARIABLE_neg_integer': '-100',
            'TESTDATA_VARIABLE_neg_float': '-1.33',
            'TESTDATA_VARIABLE_pad_integer': '001',
        },
    )
    capture = capsys.readouterr()

    assert capture.out == ''
    assert (
        capture.err
        == """
feature file test.feature will execute in total 55 iterations divided on 2 scenarios

each scenario will execute accordingly:

ident   weight  #iter  #user  description
------|-------|------|------|-------------|
001        500     50     39  scenario-1 
002          1      5      1  scenario-2 
------|-------|------|------|-------------|

"""
    )
    capsys.readouterr()
    assert ask_yes_no.call_count == 2
    args, _ = ask_yes_no.call_args_list[-1]
    assert args[0] == 'continue?'

    assert render.call_count == 5
    for _, kwargs in render.call_args_list:
        assert kwargs.get('boolean', None)
        assert kwargs.get('integer', None) == 500
        assert kwargs.get('float', None) == 1.33
        assert kwargs.get('string', None) == 'foo bar'
        assert kwargs.get('neg_integer', None) == -100
        assert kwargs.get('neg_float', None) == -1.33
        assert kwargs.get('pad_integer', None) == '001'

    mocker.patch(
        'grizzly_cli.SCENARIOS',
        [
            create_scenario(
                'scenario-1 testing a lot of stuff',
                [
                    'Given "70" users',
                ],
                [
                    'Given a user of type "RestApi" with weight "100" load testing "https://localhost"',
                    'And repeat for "500" iterationsAnd ask for value of variable "test_variable_2"',
                    'And ask for value of variable "test_variable_1"',
                ],
            ),
            create_scenario(
                'scenario-2 testing a lot more of many different things that scenario-1 does not test',
                [
                    'And ask for value of variable "bar"',
                ],
                [
                    'Given a user of type "MessageQueueUser" with weight "50" load testing "mqs://localhost"',
                    'And repeat for "750" iterations',
                    'And ask for value of variable "foo"',
                ],
            ),
            create_scenario(
                'scenario-3',
                [],
                [
                    'Given a user of type "RestApi" with weight "1" load testing "https://127.0.0.2"',
                    'And repeat for "10" iterations',
                ],
            ),
        ],
    )

    arguments = Namespace(file='integration.feature', yes=True)

    distribution_of_users_per_scenario(arguments, {})
    capture = capsys.readouterr()

    assert capture.out == ''
    assert (
        capture.err
        == """
feature file integration.feature will execute in total 1260 iterations divided on 3 scenarios

each scenario will execute accordingly:

ident   weight  #iter  #user  description                                                                         
------|-------|------|------|--------------------------------------------------------------------------------------|
001        100    500     46  scenario-1 testing a lot of stuff                                                   
002         50    750     23  scenario-2 testing a lot more of many different things that scenario-1 does not test
003          1     10      1  scenario-3                                                                          
------|-------|------|------|--------------------------------------------------------------------------------------|

"""
    )
    capsys.readouterr()
    assert ask_yes_no.call_count == 2

    mocker.patch(
        'grizzly_cli.SCENARIOS',
        [
            create_scenario(
                'scenario-1',
                [
                    'Given "1" user',
                ],
                [
                    'Given a user of type "RestApi" with weight "25" load testing "https://localhost"',
                    'And repeat for "1" iterations',
                ],
            ),
        ],
    )

    distribution_of_users_per_scenario(arguments, {})
    capture = capsys.readouterr()

    assert capture.out == ''
    assert (
        capture.err
        == """
feature file integration.feature will execute in total 1 iterations divided on 1 scenarios

each scenario will execute accordingly:

ident   weight  #iter  #user  description
------|-------|------|------|-------------|
001         25      1      1  scenario-1 
------|-------|------|------|-------------|

"""
    )
    capsys.readouterr()

    arguments = Namespace(file='integration.feature', yes=True, environment_file='environments/local.yaml')
    distribution_of_users_per_scenario(arguments, {'GRIZZLY_CONFIGURATION_FILE': 'environments/local.lock.yaml'})
    capture = capsys.readouterr()

    assert capture.out == ''
    assert (
        capture.err
        == """
feature file integration.feature will execute in total 1 iterations divided on 1 scenarios with environment file environments/local.lock.yaml

each scenario will execute accordingly:

ident   weight  #iter  #user  description
------|-------|------|------|-------------|
001         25      1      1  scenario-1 
------|-------|------|------|-------------|

"""
    )
    capsys.readouterr()

    mocker.patch(
        'grizzly_cli.SCENARIOS',
        [
            create_scenario(
                'scenario-1 testing a lot of stuff',
                [
                    'Given "4" users',
                ],
                [
                    'Given a user of type "RestApi" with weight "1" load testing "https://localhost"',
                    'And repeat for "10" iterationsAnd ask for value of variable "test_variable_2"',
                    'And ask for value of variable "test_variable_1"',
                ],
            ),
            create_scenario(
                'scenario-2 testing a lot more of many different things that scenario-1 does not test',
                [],
                [
                    'Given a user of type "MessageQueueUser" with weight "50" load testing "mqs://localhost"',
                    'And repeat for "0" iterations',
                    'And ask for value of variable "foo"',
                ],
            ),
            create_scenario(
                'scenario-3',
                [],
                [
                    'Given a user of type "RestApi" with weight "0" load testing "https://127.0.0.2"',
                    'And repeat for "10" iterations',
                ],
            ),
            create_scenario(
                'scenario-4',
                [],
                [
                    'Given a user of type "RestApi" with weight "0" load testing "https://127.0.0.2"',
                    'And repeat for "0" iterations',
                ],
            ),
        ],
    )

    arguments = Namespace(file='integration.feature', yes=True)

    with pytest.raises(ValueError) as ve:  # noqa: PT011
        distribution_of_users_per_scenario(arguments, {})
    capture = capsys.readouterr()

    assert capture.out == ''
    assert (
        capture.err
        == """
feature file integration.feature will execute in total 20 iterations divided on 4 scenarios

each scenario will execute accordingly:

ident   weight  #iter  #user  description                                                                            errors
------|-------|------|------|--------------------------------------------------------------------------------------|----------------------------------|
001          1     10      1  scenario-1 testing a lot of stuff                                                      
002         50      0      3  scenario-2 testing a lot more of many different things that scenario-1 does not test   no iterations
003          0     10      0  scenario-3                                                                             no users assigned
004          0      0      0  scenario-4                                                                             no users assigned, no iterations
------|-------|------|------|--------------------------------------------------------------------------------------|----------------------------------|
"""
    )
    assert (
        str(ve.value)
        == """                                                                                                                    ^
+-------------------------------------------------------------------------------------------------------------------+
|
+- there were errors when calculating user distribution and iterations per scenario, adjust user "weight", number of users or iterations per scenario\n"""
    )


@pytest.mark.parametrize(
    ('users', 'iterations', 'output'),
    [
        (
            6,
            13,
            """
feature file integration.feature will execute in total 28 iterations divided on 7 scenarios

each scenario will execute accordingly:

ident   weight  #iter  #user  description
------|-------|------|------|-------------|
001         50     15      5  scenario-0 
002         12      3      1  scenario-1 
003         21      5      2  scenario-2 
004          4      1      1  scenario-3 
005          6      2      1  scenario-4 
006          3      1      1  scenario-5 
007          3      1      1  scenario-6 
------|-------|------|------|-------------|

""",
        ),
        (
            12,
            20,
            """
feature file integration.feature will execute in total 43 iterations divided on 7 scenarios

each scenario will execute accordingly:

ident   weight  #iter  #user  description
------|-------|------|------|-------------|
001         33     23      6  scenario-0 
002         16      5      2  scenario-1 
003         28      8      5  scenario-2 
004          5      2      1  scenario-3 
005          8      3      2  scenario-4 
006          4      1      1  scenario-5 
007          4      1      1  scenario-6 
------|-------|------|------|-------------|

""",
        ),
        (
            18,
            31,
            """
feature file integration.feature will execute in total 66 iterations divided on 7 scenarios

each scenario will execute accordingly:

ident   weight  #iter  #user  description
------|-------|------|------|-------------|
001         25     35      6  scenario-0 
002         18      8      4  scenario-1 
003         31     13      7  scenario-2 
004          6      2      2  scenario-3 
005          9      4      3  scenario-4 
006          4      2      1  scenario-5 
007          4      2      1  scenario-6 
------|-------|------|------|-------------|

""",
        ),
        (
            24,
            49,
            """
feature file integration.feature will execute in total 105 iterations divided on 7 scenarios

each scenario will execute accordingly:

ident   weight  #iter  #user  description
------|-------|------|------|-------------|
001         20     56      6  scenario-0 
002         20     12      6  scenario-1 
003         33     21     10  scenario-2 
004          6      4      1  scenario-3 
005         10      6      3  scenario-4 
006          4      3      2  scenario-5 
007          4      3      2  scenario-6 
------|-------|------|------|-------------|

""",
        ),
        (
            30,
            58,
            """
feature file integration.feature will execute in total 124 iterations divided on 7 scenarios

each scenario will execute accordingly:

ident   weight  #iter  #user  description
------|-------|------|------|-------------|
001         16     66      6  scenario-0 
002         20     15      7  scenario-1 
003         35     24     12  scenario-2 
004          6      5      3  scenario-3 
005         10      8      4  scenario-4 
006          5      3      2  scenario-5 
007          5      3      2  scenario-6 
------|-------|------|------|-------------|

""",
        ),
        (
            30,
            21000,
            """
feature file integration.feature will execute in total 44940 iterations divided on 7 scenarios

each scenario will execute accordingly:

ident   weight  #iter  #user  description
------|-------|------|------|-------------|
001         16  23940      6  scenario-0 
002         20   5250      7  scenario-1 
003         35   8820     12  scenario-2 
004          6   1680      3  scenario-3 
005         10   2730      4  scenario-4 
006          5   1260      2  scenario-5 
007          5   1260      2  scenario-6 
------|-------|------|------|-------------|

""",
        ),
    ],
)
def test_distribution_of_users_per_scenario_advanced(capsys: CaptureFixture, mocker: MockerFixture, users: int, iterations: int, output: str) -> None:
    setup_logging()

    # all scenarios in a feature file will, at this point, have all the background steps
    # grizzly will later make sure that they are only run once
    background_steps = [
        'Given "{{ (users | int) + 6 }}" users',
        'And spawn rate is "{{ rate }}" users per second',
    ]

    mocker.patch(
        'grizzly_cli.SCENARIOS',
        [
            create_scenario(
                'scenario-0',
                background_steps,
                [
                    'Given a user of type "RestApi" with weight "{{ (6 / ((users | int) + 6) + 0.5 | int) * 100 }}" load testing "https://localhost"',
                    'And repeat for "{{ (leveranser | int) + ((((leveranser * 0.06) + 0.5) | int) or 1) + ((((leveranser * 0.08) + 0.5) | int) or 1) }}" iterations',
                ],
            ),
            create_scenario(
                'scenario-1',
                background_steps,
                [
                    'Given a user of type "RestApi" with weight "{{ (100 - ((6 / ((users | int) + 6) + 0.5 | int) * 100)) * 0.25 }}" load testing "https://localhost"',
                    'And repeat for "{{ ((leveranser * 0.25) + 0.5) | int }}" iterations',
                ],
            ),
            create_scenario(
                'scenario-2',
                background_steps,
                [
                    'Given a user of type "RestApi" with weight "{{ (100 - ((6 / ((users | int) + 6) + 0.5 | int) * 100)) * 0.42 }}" load testing "https://localhost"',
                    'And repeat for "{{ ((leveranser * 0.42) + 0.5) | int }}" iterations',
                ],
            ),
            create_scenario(
                'scenario-3',
                background_steps,
                [
                    'Given a user of type "RestApi" with weight "{{ (100 - ((6 / ((users | int) + 6) + 0.5 | int) * 100)) * 0.08 }}" load testing "https://localhost"',
                    'And repeat for "{{ ((leveranser * 0.08) + 0.5) | int }}" iterations',
                ],
            ),
            create_scenario(
                'scenario-4',
                background_steps,
                [
                    'Given a user of type "RestApi" with weight "{{ (100 - ((6 / ((users | int) + 6) + 0.5 | int) * 100)) * 0.13 }}" load testing "https://localhost"',
                    'And repeat for "{{ ((leveranser * 0.13) + 0.5) | int }}" iterations',
                ],
            ),
            create_scenario(
                'scenario-5',
                background_steps,
                [
                    'Given a user of type "RestApi" with weight "{{ (100 - ((6 / ((users | int) + 6) + 0.5 | int) * 100)) * 0.06 }}" load testing "https://localhost"',
                    'And repeat for "{{ ((leveranser * 0.06) + 0.5) | int }}" iterations',
                ],
            ),
            create_scenario(
                'scenario-6',
                background_steps,
                [
                    'Given a user of type "RestApi" with weight "{{ (100 - ((6 / ((users | int) + 6) + 0.5 | int) * 100)) * 0.06 }}" load testing "https://localhost"',
                    'And repeat for "{{ ((leveranser * 0.06) + 0.5) | int }}" iterations',
                ],
            ),
        ],
    )

    arguments = Namespace(file='integration.feature', yes=True)

    distribution_of_users_per_scenario(
        arguments,
        {
            'TESTDATA_VARIABLE_leveranser': f'{iterations}',
            'TESTDATA_VARIABLE_users': f'{users}',
            'TESTDATA_VARIABLE_rate': f'{users}',
        },
    )
    capture = capsys.readouterr()

    assert capture.out == ''
    assert capture.err == output
    capsys.readouterr()


def test_distribution_of_users_per_scenario_no_weights(capsys: CaptureFixture, mocker: MockerFixture) -> None:
    setup_logging()

    # all scenarios in a feature file will, at this point, have all the background steps
    # grizzly will later make sure that they are only run once
    background_steps = [
        'Given spawn rate is "{{ rate }}" users per second',
    ]

    mocker.patch(
        'grizzly_cli.SCENARIOS',
        [
            create_scenario(
                'scenario-0',
                background_steps,
                [
                    'Given "{{ ((((max_users * 0.7) - 0.5) | int) or 1) if max_users is defined else 1 }}" users of type "RestApi" load testing "https://localhost"',
                    'And repeat for "{{ (leveranser | int) + ((((leveranser * 0.7) + 0.5) | int) or 1) + ((((leveranser * 0.3) + 0.5) | int) or 1) }}" iterations',
                ],
            ),
            create_scenario(
                'scenario-1',
                background_steps,
                [
                    'Given "{{ ((((max_users_undefined * 0.3) - 0.5) | int) or 1) if max_users_undefined is defined else 1 }}" user of type "RestApi" load testing "https://localhost"',
                    'And repeat for "{{ ((leveranser * 0.3) + 0.5) | int }}" iterations',
                ],
            ),
        ],
    )

    arguments = Namespace(file='integration.feature', yes=True)

    distribution_of_users_per_scenario(
        arguments,
        {
            'TESTDATA_VARIABLE_leveranser': '10',
            'TESTDATA_VARIABLE_max_users': '10',
            'TESTDATA_VARIABLE_rate': '10',
        },
    )
    capture = capsys.readouterr()

    assert capture.out == ''
    assert (
        capture.err
        == """
feature file integration.feature will execute in total 23 iterations divided on 2 scenarios

each scenario will execute accordingly:

ident   #iter  #user  description
------|------|------|-------------|
001        20      6  scenario-0 
002         3      1  scenario-1 
------|------|------|-------------|

"""
    )
    capsys.readouterr()


def test_ask_yes_no(capsys: CaptureFixture, mocker: MockerFixture) -> None:
    get_input = mocker.patch('grizzly_cli.utils.get_input', side_effect=['yeah', 'n', 'y'])

    with pytest.raises(KeyboardInterrupt):
        ask_yes_no('continue?')

    capture = capsys.readouterr()
    assert capture.err == ''
    assert capture.out == 'you must answer y (yes) or n (no)\n'

    assert get_input.call_count == 2
    for args, _ in get_input.call_args_list:
        assert args[0] == 'continue? [y/n]: '
    get_input.reset_mock()

    ask_yes_no('are you sure you know what you are doing?')
    capture = capsys.readouterr()
    assert capture.err == ''
    assert capture.out == ''

    assert get_input.call_count == 1
    for args, _ in get_input.call_args_list:
        assert args[0] == 'are you sure you know what you are doing? [y/n]: '


def test_get_dependency_versions_git(mocker: MockerFixture, tmp_path_factory: TempPathFactory, capsys: CaptureFixture) -> None:  # noqa: PLR0915
    test_context = tmp_path_factory.mktemp('test_context')
    requirements_file = test_context / 'requirements.txt'

    mocker.patch('grizzly_cli.EXECUTION_CONTEXT', str(test_context))

    try:
        grizzly_versions, locust_version = get_dependency_versions(local_install=False)

        assert grizzly_versions == (None, None)
        assert locust_version is None

        requirements_file.touch()

        assert get_dependency_versions(local_install=False) == (('(unknown)', None), '(unknown)')

        capture = capsys.readouterr()
        assert capture.err == f'!! unable to find grizzly dependency in {requirements_file.absolute()}\n'
        assert capture.out == ''

        requirements_file.write_text('git+https://github.com/Biometria-se/grizzly.git@v1.5.3#egg=grizzly-loadtester')
        import subprocess

        with ExitStack() as stack:
            stack.enter_context(
                mocker.patch.context_manager(subprocess, 'check_call', side_effect=[1, 0, 0, 1, 0]),
            )
            stack.enter_context(
                mocker.patch.context_manager(
                    subprocess,
                    'check_output',
                    side_effect=[
                        subprocess.CalledProcessError(returncode=1, cmd=''),
                        'main\n',
                        'branch\n',
                        'v1.5.3\n',
                    ],
                ),
            )

            assert get_dependency_versions(local_install=False) == (('(unknown)', None), '(unknown)')

            capture = capsys.readouterr()
            assert capture.err == '!! unable to clone git repo https://github.com/Biometria-se/grizzly.git\n'
            assert capture.out == ''

            assert subprocess.check_call.call_count == 1  # type: ignore[attr-defined]
            assert subprocess.check_output.call_count == 0  # type: ignore[attr-defined]

            # git clone...
            args, kwargs = subprocess.check_call.call_args_list[0]  # type: ignore[attr-defined]
            assert len(args) == 1
            args = args[0]
            assert args[:-1] == ['git', 'clone', '--filter=blob:none', '-q', 'https://github.com/Biometria-se/grizzly.git']
            assert isinstance(args[-1], Path)
            assert args[-1].as_posix().startswith(Path(gettempdir()).as_posix())
            assert args[-1].as_posix().endswith('grizzly-loadtester_3f210f1809f6ca85ef414b2b4d450bf54353b5e0')
            assert not kwargs.get('shell', True)
            assert kwargs.get('stdout', None) == subprocess.DEVNULL
            assert kwargs.get('stderr', None) == subprocess.DEVNULL

            assert get_dependency_versions(local_install=False) == (('(unknown)', None), '(unknown)')

            capture = capsys.readouterr()
            assert capture.err == '!! unable to check branch name of HEAD in git repo https://github.com/Biometria-se/grizzly.git\n'
            assert capture.out == ''

            assert subprocess.check_call.call_count == 2  # type: ignore[attr-defined]
            assert subprocess.check_output.call_count == 1  # type: ignore[attr-defined]

            # git rev-parse...
            args, kwargs = subprocess.check_output.call_args_list[0]  # type: ignore[attr-defined]
            assert len(args) == 1
            args = args[0]
            assert args == ['git', 'rev-parse', '--abbrev-ref', 'HEAD']
            assert not kwargs.get('shell', True)
            kwarg_cwd = kwargs.get('cwd', None)
            assert isinstance(kwarg_cwd, Path)
            assert kwarg_cwd.as_posix().startswith(Path(gettempdir()).as_posix())
            assert kwarg_cwd.as_posix().endswith('grizzly-loadtester_3f210f1809f6ca85ef414b2b4d450bf54353b5e0')
            assert kwargs.get('universal_newlines', False)

            assert get_dependency_versions(local_install=True) == (('(unknown)', None), '(unknown)')

            capture = capsys.readouterr()
            assert capture.err == '!! unable to checkout branch v1.5.3 from git repo https://github.com/Biometria-se/grizzly.git\n'
            assert capture.out == ''

            assert subprocess.check_call.call_count == 4  # type: ignore[attr-defined]
            assert subprocess.check_output.call_count == 3  # type: ignore[attr-defined]

            # git checkout...
            args, kwargs = subprocess.check_call.call_args_list[-1]  # type: ignore[attr-defined]
            assert len(args) == 1
            args = args[0]
            assert args == ['git', 'checkout', '-b', 'v1.5.3', '--track', 'origin/v1.5.3']
            kwarg_cwd = kwargs.get('cwd', None)
            assert kwarg_cwd.as_posix().startswith(Path(gettempdir()).as_posix())
            assert kwarg_cwd.as_posix().endswith('grizzly-loadtester_3f210f1809f6ca85ef414b2b4d450bf54353b5e0')
            assert not kwargs.get('shell', True)
            assert kwargs.get('stdout', None) == subprocess.DEVNULL
            assert kwargs.get('stderr', None) == subprocess.DEVNULL

            with pytest.raises(FileNotFoundError):
                get_dependency_versions(local_install=False)

            capture = capsys.readouterr()
            assert capture.err == ''
            assert capture.out == ''

            assert subprocess.check_call.call_count == 5  # type: ignore[attr-defined]
            assert subprocess.check_output.call_count == 4  # type: ignore[attr-defined]

        with ExitStack() as stack:
            stack.enter_context(mocker.patch.context_manager(subprocess, 'check_call', return_value=0))
            stack.enter_context(
                mocker.patch.context_manager(subprocess, 'check_output', return_value='main\n'),
            )

            with pytest.raises(FileNotFoundError) as fne:
                get_dependency_versions(local_install=False)
            assert fne.value.errno == 2
            assert fne.value.strerror == 'No such file or directory'

            with unittest_patch(
                'grizzly_cli.utils.Path.open',
                side_effect=[
                    mock_open(read_data='git+https://github.com/Biometria-se/grizzly.git@v1.5.3#egg=grizzly-loadtester\n').return_value,
                    mock_open(read_data='').return_value,
                ],
            ) as open_mock:
                assert get_dependency_versions(local_install=False) == (('(unknown)', None), '(unknown)')

                capture = capsys.readouterr()
                assert capture.err == '!! unable to find "__version__" declaration in grizzly/__init__.py from https://github.com/Biometria-se/grizzly.git\n'
                assert capture.out == ''
                assert open_mock.call_count == 2

            with unittest_patch(
                'grizzly_cli.utils.Path.open',
                side_effect=[
                    mock_open(read_data='git+https://github.com/Biometria-se/grizzly.git@v1.5.3#egg=grizzly-loadtester\n').return_value,
                    mock_open(read_data="__version__ = '0.0.0'").return_value,
                    mock_open(read_data='').return_value,
                ],
            ) as open_mock:
                assert get_dependency_versions(local_install=False) == (('0.0.0', []), '(unknown)')

                capture = capsys.readouterr()
                assert capture.err == '!! unable to find "locust" dependency in requirements.txt from https://github.com/Biometria-se/grizzly.git\n'
                assert capture.out == ''

                assert open_mock.call_count == 3

            with unittest_patch(
                'grizzly_cli.utils.Path.open',
                side_effect=[
                    mock_open(read_data='git+https://github.com/Biometria-se/grizzly.git@v1.5.3#egg=grizzly-loadtester[dev,mq]\n').return_value,
                    mock_open(read_data="__version__ = '1.5.3'").return_value,
                    mock_open(read_data='locust').return_value,
                ],
            ) as open_mock:
                assert get_dependency_versions(local_install=False) == (('1.5.3', ['dev', 'mq']), '(unknown)')

                capture = capsys.readouterr()
                assert capture.err == '!! unable to find locust version in "locust" specified in requirements.txt from https://github.com/Biometria-se/grizzly.git\n'
                assert capture.out == ''

                assert open_mock.call_count == 3

            with unittest_patch(
                'grizzly_cli.utils.Path.open',
                side_effect=[
                    mock_open(read_data='git+https://github.com/Biometria-se/grizzly.git@v1.5.3#egg=grizzly-loadtester\n').return_value,
                    mock_open(read_data="__version__ = '1.5.3'").return_value,
                    mock_open(read_data='locust==2.2.1 \\ ').return_value,
                ],
            ) as open_mock:
                assert get_dependency_versions(local_install=False) == (('1.5.3', []), '2.2.1')

                capture = capsys.readouterr()
                assert capture.err == ''
                assert capture.out == ''

                assert open_mock.call_count == 3

            mocker.patch('grizzly_cli.utils.Path.exists', return_value=True)

            with pytest.raises(FileNotFoundError) as fne:
                get_dependency_versions(local_install=False)
            assert fne.value.errno == 2
            assert fne.value.strerror == 'No such file or directory'

            with unittest_patch(
                'grizzly_cli.utils.Path.open',
                side_effect=[
                    mock_open(read_data='git+https://github.com/Biometria-se/grizzly.git@main#egg=grizzly-loadtester\n').return_value,
                    mock_open(read_data='').return_value,
                ],
            ) as open_mock:
                assert get_dependency_versions(local_install=False) == (('(unknown)', None), '(unknown)')

                capture = capsys.readouterr()
                assert capture.err == '!! unable to find "version" declaration in setup.cfg from https://github.com/Biometria-se/grizzly.git\n'
                assert capture.out == ''
                assert open_mock.call_count == 2

            with unittest_patch(
                'grizzly_cli.utils.Path.open',
                side_effect=[
                    mock_open(read_data='git+https://github.com/Biometria-se/grizzly.git@main#egg=grizzly-loadtester[mq]\n').return_value,
                    mock_open(read_data='name = grizzly-loadtester\nversion = 2.0.0').return_value,
                    mock_open(read_data='locust==2.8.4 \\ ').return_value,
                ],
            ) as open_mock:
                assert get_dependency_versions(local_install=False) == (('2.0.0', ['mq']), '2.8.4')

                capture = capsys.readouterr()
                assert capture.err == ''
                assert capture.out == ''
                assert open_mock.call_count == 3

            with unittest_patch(
                'grizzly_cli.utils.Path.open',
                side_effect=[
                    mock_open(read_data='grizzly-loadtester @ git+https://github.com/Biometria-se/grizzly.git@main\n').return_value,
                    mock_open(read_data='').return_value,
                ],
            ) as open_mock:
                assert get_dependency_versions(local_install=False) == (('(unknown)', None), '(unknown)')

                capture = capsys.readouterr()
                assert capture.err == '!! unable to find "version" declaration in setup.cfg from https://github.com/Biometria-se/grizzly.git\n'
                assert capture.out == ''
                assert open_mock.call_count == 2

            with unittest_patch(
                'grizzly_cli.utils.Path.open',
                side_effect=[
                    mock_open(read_data='grizzly-loadtester[mq] @ git+https://github.com/Biometria-se/grizzly.git@main\n').return_value,
                    mock_open(read_data='name = grizzly-loadtester\nversion = 2.0.0').return_value,
                    mock_open(read_data='locust==2.8.4 \\ ').return_value,
                ],
            ) as open_mock:
                assert get_dependency_versions(local_install=False) == (('2.0.0', ['mq']), '2.8.4')

                capture = capsys.readouterr()
                assert capture.err == ''
                assert capture.out == ''
                assert open_mock.call_count == 3

            with unittest_patch(
                'grizzly_cli.utils.Path.open',
                side_effect=[
                    mock_open(read_data='grizzly-loadtester[mq] % git+https://github.com/Biometria-se/grizzly.git@main\n').return_value,
                ],
            ) as open_mock:
                assert get_dependency_versions(local_install=False) == (('(unknown)', None), '(unknown)')

                capture = capsys.readouterr()
                assert capture.err == f'!! unable to find properly formatted grizzly dependency in {requirements_file}\n'
                assert capture.out == ''
                assert open_mock.call_count == 1
    finally:
        rm_rf(test_context)


@pytest.mark.filterwarnings('ignore:Creating a LegacyVersion has been deprecated')
def test_get_dependency_versions_pypi(mocker: MockerFixture, tmp_path_factory: TempPathFactory, capsys: CaptureFixture, requests_mock: RequestsMocker) -> None:  # noqa: PLR0915
    test_context = tmp_path_factory.mktemp('test_context')
    requirements_file = test_context / 'requirements.txt'

    mocker.patch('grizzly_cli.EXECUTION_CONTEXT', str(test_context))

    try:
        grizzly_versions, locust_version = get_dependency_versions(local_install=False)

        assert grizzly_versions == (None, None)
        assert locust_version is None

        requirements_file.touch()

        assert get_dependency_versions(local_install=False) == (('(unknown)', None), '(unknown)')

        capture = capsys.readouterr()
        assert capture.err == f'!! unable to find grizzly dependency in {requirements_file.absolute()}\n'
        assert capture.out == ''

        requirements_file.write_text('grizzly-loadtester')

        requests_mock.register_uri('GET', 'https://pypi.org/pypi/grizzly-loadtester/json', status_code=404)

        assert get_dependency_versions(local_install=False) == (('(unknown)', None), '(unknown)')

        capture = capsys.readouterr()
        assert capture.err == '!! unable to get grizzly package information from https://pypi.org/pypi/grizzly-loadtester/json (404)\n'
        assert capture.out == ''

        requests_mock.register_uri('GET', 'https://pypi.org/pypi/grizzly-loadtester/json', status_code=200, text='{"info": {"version": "1.1.1"}}')
        requests_mock.register_uri('GET', 'https://pypi.org/pypi/grizzly-loadtester/1.1.1/json', status_code=400)

        assert get_dependency_versions(local_install=False) == (('1.1.1', []), '(unknown)')

        capture = capsys.readouterr()
        assert capture.err == '!! unable to get grizzly 1.1.1 package information from https://pypi.org/pypi/grizzly-loadtester/1.1.1/json (400)\n'
        assert capture.out == ''

        requests_mock.register_uri('GET', 'https://pypi.org/pypi/grizzly-loadtester/1.1.1/json', status_code=200, text='{"info": {"requires_dist": []}}')

        assert get_dependency_versions(local_install=True) == (('1.1.1', []), '(unknown)')

        capture = capsys.readouterr()
        assert capture.err == '!! could not find "locust" in requires_dist information for grizzly-loadtester 1.1.1\n'
        assert capture.out == ''

        requests_mock.register_uri('GET', 'https://pypi.org/pypi/grizzly-loadtester/1.1.1/json', status_code=200, text='{"info": {"requires_dist": ["requests", "locust"]}}')

        requirements_file.unlink()
        requirements_file.write_text('grizzly-loadtester[dev,mq]')

        actual_dependency_versions = get_dependency_versions(local_install=True)

        assert actual_dependency_versions == (('1.1.1', ['dev', 'mq']), '(unknown)')

        capture = capsys.readouterr()
        assert capture.err == '!! unable to find locust version in "locust" specified in pypi for grizzly-loadtester 1.1.1\n'
        assert capture.out == ''

        requests_mock.register_uri('GET', 'https://pypi.org/pypi/grizzly-loadtester/1.1.1/json', status_code=200, text='{"info": {"requires_dist": ["locust (==2.8.5)"]}}')

        assert get_dependency_versions(local_install=False) == (('1.1.1', ['dev', 'mq']), '2.8.5')

        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == ''

        requirements_file.unlink()
        requirements_file.write_text('grizzly-loadtester[mq]==1.4.0')

        requests_mock.register_uri('GET', 'https://pypi.org/pypi/grizzly-loadtester/json', status_code=200, text='{"releases": {"1.3.0": [], "1.5.0": []}}')

        assert get_dependency_versions(local_install=False) == (('(unknown)', None), '(unknown)')

        capture = capsys.readouterr()
        assert capture.err == '!! could not resolve grizzly-loadtester[mq]==1.4.0 to one specific version available at pypi\n'
        assert capture.out == ''

        requirements_file.unlink()
        requirements_file.write_text('grizzly-loadtester[mq]>1.3.0,<1.5.0')

        requests_mock.register_uri(
            'GET',
            'https://pypi.org/pypi/grizzly-loadtester/json',
            status_code=200,
            text='{"releases": {"1.3.0": [], "1.4.0": [], "1.5.0": [], "foobar": []}}',
        )
        requests_mock.register_uri('GET', 'https://pypi.org/pypi/grizzly-loadtester/1.4.0/json', status_code=200, text='{"info": {"requires_dist": ["locust (==1.0.0)"]}}')

        assert get_dependency_versions(local_install=False) == (('1.4.0', ['mq']), '1.0.0')

        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == ''

        requirements_file.unlink()
        requirements_file.write_text('grizzly-loadtester[mq]<=1.5.0')

        requests_mock.register_uri(
            'GET',
            'https://pypi.org/pypi/grizzly-loadtester/json',
            status_code=200,
            text='{"releases": {"1.4.20": [], "1.5.0": [], "1.5.1": []}}',
        )
        requests_mock.register_uri('GET', 'https://pypi.org/pypi/grizzly-loadtester/1.5.0/json', status_code=200, text='{"info": {"requires_dist": ["locust (==1.1.1)"]}}')

        assert get_dependency_versions(local_install=False) == (('1.5.0', ['mq']), '1.1.1')

        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == ''

        local_install_requirements_file = test_context / 'a' / 'b' / 'requirements.txt'
        local_install_requirements_file.parent.mkdir(parents=True)
        local_install_requirements_file.write_text('grizzly-loadtester[mq]>1.3.0,<1.5.0')

        requests_mock.register_uri(
            'GET',
            'https://pypi.org/pypi/grizzly-loadtester/json',
            status_code=200,
            text='{"releases": {"1.3.0": [], "1.4.0": [], "1.5.0": [], "foobar": []}}',
        )
        requests_mock.register_uri('GET', 'https://pypi.org/pypi/grizzly-loadtester/1.4.0/json', status_code=200, text='{"info": {"requires_dist": ["locust (==1.0.0)"]}}')

        assert get_dependency_versions(local_install='a/b/') == (('1.4.0', ['mq']), '1.0.0')
        assert get_dependency_versions(local_install='a/b/requirements.txt') == (('1.4.0', ['mq']), '1.0.0')
    finally:
        rm_rf(test_context)


def test_requirements(capsys: CaptureFixture, tmp_path_factory: TempPathFactory) -> None:
    test_context = tmp_path_factory.mktemp('test_context')
    requirements_file = test_context / 'requirements.txt'

    def wrapped_test(_args: Namespace) -> int:
        return 1337

    try:
        assert not requirements_file.exists()

        wrapped = requirements(test_context.as_posix())(wrapped_test)
        assert getattr(wrapped, '__wrapped__', None) is wrapped_test
        assert getattr(getattr(wrapped, '__wrapped__'), '__value__') == test_context.as_posix()  # noqa: B009

        assert wrapped(Namespace()) == 1337

        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == ''
        assert not requirements_file.exists()

    finally:
        rm_rf(test_context)



"""Tests for grizzly_cli.init."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from grizzly_cli.__main__ import _parse_arguments
from grizzly_cli.init import init, tree

from test_cli.helpers import rm_rf

if TYPE_CHECKING:  # pragma: no cover
    from _pytest.capture import CaptureFixture
    from _pytest.tmpdir import TempPathFactory
    from pytest_mock import MockerFixture


def test_tree(tmp_path_factory: TempPathFactory) -> None:
    test_context = tmp_path_factory.mktemp('test_context')

    (test_context / 'a' / 'b' / 'c').mkdir(parents=True)

    (test_context / 'a' / 'file-a1.txt').touch()
    (test_context / 'a' / 'file-a2.txt').touch()

    (test_context / 'a' / 'b' / 'file-b1.txt').touch()

    (test_context / 'a' / 'b' / 'c' / 'file-c1.txt').touch()
    (test_context / 'a' / 'b' / 'c' / 'file-c2.txt').touch()

    (test_context / 'root.yaml').touch()

    try:
        assert (
            '\n'.join(list(tree(test_context)))
            == """├── a
│   ├── b
│   │   ├── c
│   │   │   ├── file-c1.txt
│   │   │   └── file-c2.txt
│   │   └── file-b1.txt
│   ├── file-a1.txt
│   └── file-a2.txt
└── root.yaml"""
        )
    finally:
        rm_rf(test_context)


def test_init(tmp_path_factory: TempPathFactory, capsys: CaptureFixture, mocker: MockerFixture) -> None:  # noqa: PLR0915
    test_context = tmp_path_factory.mktemp('test_context')

    test_existing = test_context / 'foobar'
    test_existing.mkdir()

    mocker.patch('grizzly_cli.init.EXECUTION_CONTEXT', str(test_context))

    try:
        sys.argv = ['grizzly-cli', 'init', 'foobar']
        arguments = _parse_arguments()

        assert init(arguments) == 1

        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == f'"foobar" already exists in {test_context}\n'

        (test_existing / 'environments').mkdir()
        (test_existing / 'features').mkdir()
        (test_existing / 'requirements.txt').touch()

        mocker.patch('grizzly_cli.init.EXECUTION_CONTEXT', str(test_existing))

        assert init(arguments) == 1

        capture = capsys.readouterr()
        assert capture.err == ''
        assert (
            capture.out
            == f"""oops, looks like you are already in a grizzly project directory

{test_existing}
├── environments
├── features
└── requirements.txt
"""
        )

        rm_rf(test_existing)

        question_mock = mocker.patch('grizzly_cli.init.ask_yes_no', side_effect=[None] * 4)
        mocker.patch('grizzly_cli.init.EXECUTION_CONTEXT', str(test_context))

        assert init(arguments) == 0

        assert question_mock.call_count == 1
        args, _ = question_mock.call_args_list[-1]
        assert (
            args[0]
            == """the following structure will be created:

    foobar
    ├── environments
    │   └── foobar.yaml
    ├── features
    │   ├── environment.py
    │   ├── steps
    │   │   └── steps.py
    │   ├── foobar.feature
    │   └── requests
    └── requirements.txt

do you want to create grizzly project "foobar"?"""
        )

        capture = capsys.readouterr()
        assert capture.err == ''
        assert (
            capture.out
            == """successfully created project "foobar", with the following options:
  • without IBM MQ support
  • latest grizzly version
"""
        )

        template_root = test_context / 'foobar'
        assert template_root.is_dir()
        assert (template_root / 'environments').is_dir()
        environments_file = template_root / 'environments' / 'foobar.yaml'
        assert environments_file.is_file()
        assert (
            environments_file.read_text()
            == """configuration:
  template:
    host: https://localhost
"""
        )

        assert (template_root / 'features').is_dir()
        feature_file = template_root / 'features' / 'foobar.feature'
        assert feature_file.is_file()
        assert (
            feature_file.read_text()
            == """Feature: Template feature file
  Scenario: Template scenario
    Given a user of type "RestApi" with weight "1" load testing "$conf::template.host"
"""
        )

        environment_file = template_root / 'features' / 'environment.py'
        assert environment_file.is_file()
        assert environment_file.read_text() == 'from grizzly.behave import *\n\n'

        assert (template_root / 'features' / 'requests').is_dir()

        assert (template_root / 'features' / 'steps').is_dir()

        steps_file = template_root / 'features' / 'steps' / 'steps.py'
        assert steps_file.is_file()
        assert steps_file.read_text() == 'from grizzly.steps import *\n\n'

        requirements_file = template_root / 'requirements.txt'
        assert requirements_file.is_file()
        assert requirements_file.read_text() == 'grizzly-loadtester\n'

        created_structure = '\n'.join(list(tree(template_root)))
        assert (
            created_structure
            == """├── environments
│   └── foobar.yaml
├── features
│   ├── environment.py
│   ├── foobar.feature
│   ├── requests
│   └── steps
│       └── steps.py
└── requirements.txt"""
        )

        rm_rf(template_root)

        capsys.readouterr()

        sys.argv = ['grizzly-cli', 'init', 'foobar', '--with-mq']
        arguments = _parse_arguments()

        assert init(arguments) == 0

        capture = capsys.readouterr()
        assert capture.err == ''
        assert (
            capture.out
            == """successfully created project "foobar", with the following options:
  • with IBM MQ support
  • latest grizzly version
"""
        )
        requirements_file = template_root / 'requirements.txt'
        assert requirements_file.is_file()
        assert requirements_file.read_text() == 'grizzly-loadtester[mq]\n'

        rm_rf(template_root)

        sys.argv = ['grizzly-cli', 'init', 'foobar', '--grizzly-version', '1.2.4']
        arguments = _parse_arguments()

        assert init(arguments) == 0

        capture = capsys.readouterr()
        assert capture.err == ''
        assert (
            capture.out
            == """successfully created project "foobar", with the following options:
  • without IBM MQ support
  • pinned to grizzly version 1.2.4
"""
        )
        requirements_file = template_root / 'requirements.txt'
        assert requirements_file.is_file()
        assert requirements_file.read_text() == 'grizzly-loadtester==1.2.4\n'

        assert (template_root / 'features').is_dir()
        feature_file = template_root / 'features' / 'foobar.feature'
        assert feature_file.is_file()
        assert (
            feature_file.read_text()
            == """Feature: Template feature file
  Scenario: Template scenario
    Given a user of type "RestApi" with weight "1" load testing "$conf::template.host"
"""
        )

        environment_file = template_root / 'features' / 'environment.py'
        assert environment_file.is_file()
        assert environment_file.read_text() == 'from grizzly.environment import *\n\n'

        rm_rf(template_root)

        sys.argv = ['grizzly-cli', 'init', 'foobar', '--with-mq', '--grizzly-version', '1.5.0']
        arguments = _parse_arguments()

        assert init(arguments) == 0

        capture = capsys.readouterr()
        assert capture.err == ''
        assert (
            capture.out
            == """successfully created project "foobar", with the following options:
  • with IBM MQ support
  • pinned to grizzly version 1.5.0
"""
        )
        requirements_file = template_root / 'requirements.txt'
        assert requirements_file.is_file()
        assert requirements_file.read_text() == 'grizzly-loadtester[mq]==1.5.0\n'

        rm_rf(template_root)

        sys.argv = ['grizzly-cli', 'init', 'foobar', '--yes']
        arguments = _parse_arguments()

        assert init(arguments) == 0

        capture = capsys.readouterr()
        assert capture.err == ''
        assert (
            capture.out
            == """the following structure will be created:

    foobar
    ├── environments
    │   └── foobar.yaml
    ├── features
    │   ├── environment.py
    │   ├── steps
    │   │   └── steps.py
    │   ├── foobar.feature
    │   └── requests
    └── requirements.txt

successfully created project "foobar", with the following options:
  • without IBM MQ support
  • latest grizzly version
"""
        )
        requirements_file = template_root / 'requirements.txt'
        assert requirements_file.is_file()
        assert requirements_file.read_text() == 'grizzly-loadtester\n'
    finally:
        rm_rf(test_context)



"""Tests for grizzly_cli.utils.configuration."""

from __future__ import annotations

from base64 import b64encode
from pathlib import Path
from platform import system
from typing import TYPE_CHECKING
from unittest.mock import MagicMock

import pytest
from azure.core.exceptions import ClientAuthenticationError, ServiceRequestError
from azure.identity import ChainedTokenCredential
from azure.keyvault.secrets import KeyVaultSecret, SecretClient, SecretProperties
from grizzly_cli.utils import chunker, setup_logging
from grizzly_cli.utils.configuration import (
    ScenarioTag,
    _get_metadata,
    _import_files,
    _write_file,
    get_context_root,
    get_keyvault_client,
    load_configuration,
    load_configuration_file,
    load_configuration_keyvault,
)

from test_cli.helpers import ANY, cwd, rm_rf

if TYPE_CHECKING:
    from _pytest.tmpdir import TempPathFactory
    from pytest_mock import MockerFixture


def create_secret_property(name: str, content_type: str | None = None) -> SecretProperties:
    return SecretProperties(vault_id=f'https://grizzly.keyvault.com/foo/{name}', content_type=content_type)


def create_secret(name: str, value: str | None, content_type: str | None = None) -> KeyVaultSecret:
    return KeyVaultSecret(create_secret_property(name, content_type), value)


def mock_keyvault(client: MagicMock, keyvault: dict[str, tuple[str | None, str | None]]) -> None:
    secret_properties: list[SecretProperties] = []
    secrets: list[KeyVaultSecret] = []

    for key, (value, content_type) in keyvault.items():
        secret = create_secret(key, value, content_type)
        secret_properties.append(secret.properties)
        if key.startswith('grizzly--'):
            secrets.append(secret)

    client.reset_mock()
    client.list_properties_of_secrets.return_value = secret_properties
    client.get_secret.side_effect = secrets


class TestScenarioTag:
    def test_get_scenario_text(self, tmp_path_factory: TempPathFactory) -> None:
        test_context = tmp_path_factory.mktemp('context')
        test_feature = test_context / 'test.feature'

        try:
            test_feature.write_text("""Feature: test
    Background: common
        Then some steps here
        And other useful stuff

    Scenario: first
        \"\"\"
        this is just some comments
        for the scenario, that spans
        multiple lines.
        \"\"\"
        Given the first scenario
        And it's steps

    Scenario: second
        Given the second scenario

        # <!-- comment -->
        And it's steps
        \"\"\"
        step text
        \"\"\"

    Scenario: third
        \"\"\"
        this is just some comments
        for the scenario, that spans
        multiple lines.
        \"\"\"
        Given the third scenario
        And it's steps
        | foo | bar |
        | bar | foo |

        And one more step
""")

            assert (
                ScenarioTag.get_scenario_text('first', test_feature)
                == """Given the first scenario
        And it's steps"""
            )

            assert (
                ScenarioTag.get_scenario_text('second', test_feature)
                == """Given the second scenario

        # <!-- comment -->
        And it's steps
        \"\"\"
        step text
        \"\"\""""
            )

            assert (
                ScenarioTag.get_scenario_text('third', test_feature)
                == """Given the third scenario
        And it's steps
        | foo | bar |
        | bar | foo |

        And one more step"""
            )
        finally:
            rm_rf(test_context)


def test_get_keyvault_client(mocker: MockerFixture) -> None:
    secrets_client_mock = mocker.patch('grizzly_cli.utils.configuration.SecretClient', return_value=MagicMock(spec=SecretClient))

    assert get_keyvault_client('https://grizzly.keyvault.azure.com') == secrets_client_mock.return_value

    secrets_client_mock.assert_called_once_with(vault_url='https://grizzly.keyvault.azure.com', credential=ANY(ChainedTokenCredential))
    secrets_client_mock.reset_mock()


def test__get_metadata() -> None:
    assert _get_metadata('hello', 'file') is None
    assert _get_metadata('file:foo.bar.txt', 'file') == 'foo.bar.txt'
    assert _get_metadata('foobar,file:test.txt,noconf', 'file') == 'test.txt'


def test_get_context_root(tmp_path_factory: TempPathFactory, mocker: MockerFixture) -> None:
    test_context = tmp_path_factory.mktemp('test_context')

    test_path = Path(test_context.as_posix())
    mocker.patch('grizzly_cli.utils.configuration.Path.cwd', return_value=test_path)
    rglob_mock = mocker.patch('grizzly_cli.utils.configuration.Path.rglob', return_value=None)

    try:
        actual_context_root = test_context / 'features' / 'environment.py'
        rglob_mock.return_value = iter([actual_context_root])

        assert get_context_root().as_posix() == actual_context_root.parent.as_posix()

        fake_context_root1 = test_context / 'asomething' / '.env' / 'environment.py'
        fake_context_root2 = test_context / 'asomething' / '.env' / 'lib' / 'python3.1337' / 'site-packages' / 'library' / 'environment.py'

        rglob_mock.return_value = iter([fake_context_root1, fake_context_root2, actual_context_root])

        assert get_context_root().as_posix() == actual_context_root.parent.as_posix()

        rglob_mock.return_value = iter([])

        with pytest.raises(ValueError, match='context root not found, are you in a grizzly project?'):
            get_context_root()
    finally:
        rm_rf(test_context)


def test__write_file(tmp_path_factory: TempPathFactory) -> None:
    test_context = tmp_path_factory.mktemp('test_context')

    try:
        assert _write_file(test_context, 'file:foo/bar.txt', b64encode(b'foo bar').decode('utf-8')) == 'files/foo/bar.txt'
        f = test_context / 'files' / 'foo' / 'bar.txt'
        assert f.read_text() == 'foo bar'

        if system() != 'Windows':
            assert f.parent.stat().st_mode & 0x000FFF == 0o700
            assert f.stat().st_mode & 0x000FFF == 0o600

        f = test_context / 'files' / 'foobar.txt'
        f.touch()

        chunks = chunker(b64encode(b'foobarfoobarfoobar').decode('utf-8'), 8)
        number_of_chunks = len(chunks)
        processed_chunks: list[str] = []

        for index, chunk in enumerate(chunks):
            content_type = f'file:foobar.txt,chunk:{index},chunks:{number_of_chunks}'
            processed_chunks.append(chunk)

            assert _write_file(test_context, content_type, chunk) == 'files/foobar.txt'
            f = test_context / 'files' / 'foobar.txt'
            expected = ''.join(processed_chunks) if index < number_of_chunks - 1 else 'foobarfoobarfoobar'

            assert f.read_text() == expected

        if system() != 'Windows':
            assert f.parent.stat().st_mode & 0x000FFF == 0o700
            assert f.stat().st_mode & 0x000FFF == 0o600

        with pytest.raises(ValueError, match='could not find `file:` in content type'):
            _write_file(test_context, 'noconf,chunk:0,chunks:2', 'foobar')
    finally:
        rm_rf(test_context)


def test__import_files(mocker: MockerFixture, tmp_path_factory: TempPathFactory) -> None:
    test_context = tmp_path_factory.mktemp('test_context')

    client_mock = mocker.MagicMock(spec=SecretClient)
    client_mock._vault_url = 'https://grizzly.keyvault.com'

    secret = create_secret('grizzly--local--foo-bar', None)

    try:
        with pytest.raises(ValueError, match='secret grizzly--local--foo-bar has no value'):
            _import_files(client_mock, test_context, secret)

        secret = create_secret('grizzly--local--foo-bar', 'grizzly--local--bar-foo')
        mock_keyvault(
            client_mock,
            {
                'grizzly--local-bar-foo': ('barfoo', None),
            },
        )

        with pytest.raises(ValueError, match='secret grizzly--local--bar-foo has no content type'):
            _import_files(client_mock, test_context, secret)

        secret = create_secret('grizzly--local--foo-bar', 'grizzly--local--bar-foo')
        mock_keyvault(
            client_mock,
            {
                'grizzly--local--bar-foo': (None, 'file:foobar.txt,chunk:0,chunks:2'),
            },
        )

        with pytest.raises(ValueError, match='secret grizzly--local--bar-foo has no value'):
            _import_files(client_mock, test_context, secret)

        secret = create_secret('grizzly--local--foo-bar', 'grizzly--local--bar-foo')
        mock_keyvault(
            client_mock,
            {
                'grizzly--local--bar-foo': (b64encode(b'foobar').decode('utf-8'), 'file:foobar.txt'),
            },
        )

        f = test_context / 'files' / 'foobar.txt'

        assert _import_files(client_mock, test_context, secret) == 'files/foobar.txt'
        assert f.read_text() == 'foobar'

        secret = create_secret('grizzly--local--foo-bar', 'grizzly--local--bar-foo')
        mock_keyvault(
            client_mock,
            {
                'grizzly--local--bar-foo': ('grizzly--local--foo-baz,grizzly--local--baz-foo', 'files'),
                'grizzly--local--foo-baz': (b64encode(b'foo baz').decode('utf-8'), 'file:foo/baz.txt,noconf'),
                'grizzly--local--baz-foo': (b64encode(b'baz foo').decode('utf-8'), 'file:baz/foo.txt,noconf'),
            },
        )

        assert _import_files(client_mock, test_context, secret) == 'files/baz/foo.txt'

        f = test_context / 'files' / 'foo' / 'baz.txt'
        assert f.read_text() == 'foo baz'

        f = test_context / 'files' / 'baz' / 'foo.txt'
        assert f.read_text() == 'baz foo'
    finally:
        rm_rf(test_context)


def test_load_configuration(mocker: MockerFixture, tmp_path_factory: TempPathFactory) -> None:
    test_context = tmp_path_factory.mktemp('test_context')

    load_configuration_keyvault_mock = mocker.patch('grizzly_cli.utils.configuration.load_configuration_keyvault', return_value=({}, 0))
    context_root = test_context.parent

    mocker.patch('grizzly_cli.utils.configuration.get_context_root', return_value=context_root)

    try:
        env_file_local = test_context / 'local.yaml'

        env_file_local.write_text("""configuration:
    authentication:
        admin:
            username: administrator
            password: hunter
""")

        env_file_lock = load_configuration(env_file_local)
        assert env_file_lock.as_posix() == Path.joinpath(test_context, 'local.lock.yaml').as_posix()

        if system() != 'Windows':
            assert env_file_lock.stat().st_mode & 0x000FFF == 0o600

        assert env_file_lock.read_text() == env_file_local.read_text()
        load_configuration_keyvault_mock.assert_not_called()

        with cwd(test_context):
            env_file_lock = load_configuration(Path('local.yaml'))
            assert env_file_lock.as_posix() == 'local.lock.yaml'

            assert env_file_lock.read_text() == env_file_local.read_text()
            load_configuration_keyvault_mock.assert_not_called()

        env_file_local.write_text("""configuration:
    keyvault: https://grizzly.keyvault.azure.com
    authentication:
        admin:
            username: administrator
            password: hunter
""")

        env_file_lock = load_configuration(env_file_local)
        assert env_file_lock.as_posix() == Path.joinpath(test_context, 'local.lock.yaml').as_posix()

        assert env_file_lock.read_text() == env_file_local.read_text()
        load_configuration_keyvault_mock.assert_called_once_with(ANY(SecretClient), 'local', context_root, filter_keys=None)
        load_configuration_keyvault_mock.reset_mock()

        env_file_local.write_text("""configuration:
    env: test
    keyvault: https://grizzly.keyvault.azure.com
    authentication:
        admin:
            username: administrator
            password: hunter
""")

        env_file_lock = load_configuration(env_file_local)
        assert env_file_lock.as_posix() == Path.joinpath(test_context, 'local.lock.yaml').as_posix()

        assert env_file_lock.read_text() == env_file_local.read_text()
        load_configuration_keyvault_mock.assert_called_once_with(ANY(SecretClient), 'test', context_root, filter_keys=None)
        load_configuration_keyvault_mock.reset_mock()

        dummy_file = test_context / 'dummy.txt'
        with pytest.raises(ValueError, match='dummy.txt does not exist'):
            load_configuration(dummy_file)

        dummy_file.touch()

        with pytest.raises(ValueError, match='configuration file must have file extension yml or yaml'):
            load_configuration(dummy_file)
    finally:
        rm_rf(test_context)


def test_load_configuration_file(tmp_path_factory: TempPathFactory) -> None:
    test_context = tmp_path_factory.mktemp('test_context')

    try:
        env_file_base = test_context / 'base.yaml'

        env_file_base.write_text("""configuration:
    authentication:
        admin:
            username: administrator
            password: hunter
""")

        assert load_configuration_file(env_file_base) == {
            'configuration': {
                'authentication': {
                    'admin': {
                        'username': 'administrator',
                        'password': 'hunter',
                    },
                },
            },
        }

        env_file_local = test_context / 'local.yaml'
        env_file_local.write_text("""{% merge "./base.yaml" %}
configuration:
    authentication:
        admin:
            username: root
            password: null
    logging:
        level: DEBUG
        max_size: 10000
""")
        assert load_configuration_file(env_file_local) == {
            'configuration': {
                'authentication': {
                    'admin': {
                        'username': 'administrator',
                        'password': 'hunter',
                    },
                },
                'logging': {
                    'level': 'DEBUG',
                    'max_size': 10000,
                },
            },
        }

    finally:
        rm_rf(test_context)


def test_load_configuration_keyvault(mocker: MockerFixture, tmp_path_factory: TempPathFactory) -> None:
    setup_logging()

    context_root = tmp_path_factory.mktemp('test_context')

    client_mock = mocker.MagicMock(spec=SecretClient)
    client_mock._vault_url = 'https://grizzly.keyvault.com'

    try:
        # <!-- no matching secrets
        client_mock.list_properties_of_secrets.return_value = [create_secret_property('grizzly--none--foo-bar')]

        assert load_configuration_keyvault(client_mock, 'local', context_root, filter_keys=None) == ({}, 0)

        client_mock.list_properties_of_secrets.assert_called_once_with()
        client_mock.get_secret.assert_not_called()
        # // -->

        # <!-- one global, one environment specific
        mock_keyvault(
            client_mock,
            {
                'some--random--key': ('rando', None),
                'grizzly--global--foo-bar': ('foobar', None),
                'grizzly--local--bar-foo-bar-foo': ('barfoo', None),
                'grizzly--remote--bar-foo-bar-foo': ('foobar', None),
            },
        )
        keyvault_configuration, imported_from_keyvault = load_configuration_keyvault(client_mock, 'local', context_root, filter_keys=None)
        assert keyvault_configuration == {
            'foo': {
                'bar': 'foobar',
            },
            'bar': {
                'foo': {
                    'bar': {
                        'foo': 'barfoo',
                    },
                },
            },
        }
        assert imported_from_keyvault == 2
        # // -->

        # <!-- one global, one environment specific, filter on one
        mock_keyvault(
            client_mock,
            {
                'some--random--key': ('rando', None),
                'grizzly--global--foo-bar': ('foobar', None),
                'grizzly--local--bar-foo-bar-foo': ('barfoo', None),
                'grizzly--remote--bar-foo-bar-foo': ('foobar', None),
            },
        )
        keyvault_configuration, imported_from_keyvault = load_configuration_keyvault(client_mock, 'local', context_root, filter_keys=['bar.foo.bar.foo'])
        assert keyvault_configuration == {
            'bar': {
                'foo': {
                    'bar': {
                        'foo': 'barfoo',
                    },
                },
            },
        }
        assert imported_from_keyvault == 1
        # // -->

        # <!-- ClientAuthenticationError
        client_mock.list_properties_of_secrets.side_effect = [ClientAuthenticationError]

        with pytest.raises(ClientAuthenticationError):
            load_configuration_keyvault(client_mock, 'local', context_root, filter_keys=None)
        # // -->

        # <!-- ServiceRequestError
        client_mock.list_properties_of_secrets.side_effect = [ServiceRequestError(message='error')]

        with pytest.raises(ServiceRequestError):
            load_configuration_keyvault(client_mock, 'local', context_root, filter_keys=None)
        # // -->
    finally:
        rm_rf(context_root)



"""Tests for grizzly-cli * run."""

from __future__ import annotations

import sys
from os import pathsep
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING

import pytest
import yaml

from test_cli.helpers import rm_rf, run_command

if TYPE_CHECKING:
    from test_cli.fixtures import End2EndFixture


def prepare_example_project(e2e_fixture: End2EndFixture) -> Path:
    example_root = e2e_fixture.root / 'grizzly-example'

    example_root.mkdir()

    rc, output = run_command(
        [
            'git',
            'init',
        ],
        cwd=example_root,
    )

    try:
        assert rc == 0
    except AssertionError:
        print(''.join(output))
        raise

    rc, output = run_command(
        [
            'git',
            'remote',
            'add',
            '-f',
            'origin',
            'https://github.com/Biometria-se/grizzly.git',
        ],
        cwd=example_root,
    )

    try:
        assert rc == 0
    except AssertionError:
        print(''.join(output))
        raise

    rc, output = run_command(
        [
            'git',
            'sparse-checkout',
            'init',
        ],
        cwd=example_root,
    )

    try:
        assert rc == 0
    except AssertionError:
        print(''.join(output))
        raise

    rc, output = run_command(
        [
            'git',
            'sparse-checkout',
            'set',
            'example',
        ],
        cwd=example_root,
    )

    try:
        assert rc == 0
    except AssertionError:
        print(''.join(output))
        raise

    rc, output = run_command(
        [
            'git',
            'pull',
            'origin',
            'main',
        ],
        cwd=example_root,
    )

    try:
        assert rc == 0
    except AssertionError:
        print(''.join(output))
        raise

    rm_rf(example_root / '.git')

    return example_root / 'example'


def validate_result(rc: int, result: str, example_root: Path) -> None:
    assert rc == 0
    assert 'ERROR' not in result
    assert 'WARNING' not in result
    assert '1 feature passed, 0 failed, 0 skipped' in result
    assert '3 scenarios passed, 0 failed, 0 skipped' in result
    assert 'steps passed, 0 failed, 0 skipped' in result

    assert 'ident   iter  status   description' in result
    assert '001      2/2  passed   dog facts api' in result
    assert '002      1/1  passed   cat facts api' in result
    assert '003      1/1  passed   book api' in result
    assert '------|-----|--------|---------------|' in result

    assert 'executing custom.User.request for 002 get-cat-facts and /facts?limit=' in result

    assert 'sending "client_server" from CLIENT' in result
    assert 'received from CLIENT' in result
    assert "AtomicCustomVariable.foobar='foobar'" in result

    assert 'compose.yaml: `version` is obsolete' not in result

    log_file_result = (example_root / 'test_run.log').read_text()

    # problems with a locust DEBUG log message containing ERROR in the message on macos-latest
    if sys.platform == 'darwin':
        output = [line for line in log_file_result.split('\n') if 'ERROR' not in line and 'DEBUG' not in line]
        log_file_result = '\n'.join(output)

    if sys.version_info >= (3, 12):
        result = result.replace('\r', '\n')

    assert log_file_result == result


def install_dependencies(e2e_fixture: End2EndFixture, example_root: Path) -> None:
    if e2e_fixture._distributed:
        command = ['grizzly-cli', 'dist', '--project-name', e2e_fixture.root.name, 'build', '--no-cache']
        rc, output = run_command(
            command,
            cwd=example_root,
            env=e2e_fixture._env,
        )
        try:
            assert rc == 0
        except AssertionError:
            with e2e_fixture.log_file.open('a+') as fd:
                fd.write(''.join(output))
            raise
    else:
        command = ['uv', 'sync', '--active', '--locked', '--package', 'grizzly-loadtester']
        repo_root_path = (Path(__file__).parent / '..' / '..' / '..' / '..').resolve()

        rc, output = run_command(
            command,
            cwd=repo_root_path,
            env=e2e_fixture._env,
        )

        try:
            assert rc == 0
        except AssertionError:
            with e2e_fixture.log_file.open('a+') as fd:
                fd.write(''.join(output))
            raise


def test_e2e_run_example(e2e_fixture: End2EndFixture) -> None:
    if sys.platform == 'win32' and e2e_fixture._distributed:
        pytest.skip('windows github runners do not support running linux containers')

    result: str | None = None

    try:
        example_root = prepare_example_project(e2e_fixture)

        with (example_root / 'features' / 'steps' / 'steps.py').open('a') as fd:
            fd.write(e2e_fixture.start_webserver_step_impl(e2e_fixture.webserver_port))

        e2e_fixture.inject_webserver_module(example_root)

        with (example_root / 'environments' / 'example.yaml').open('r') as env_yaml_file:
            env_conf = yaml.full_load(env_yaml_file)

            for name in ['dog', 'cat', 'book']:
                env_conf['configuration']['facts'][name]['host'] = f'http://{e2e_fixture.host}'

        feature_file = Path.joinpath(Path('features'), 'example.feature')
        feature_file_path = example_root / 'features' / 'example.feature'
        feature_file_contents = feature_file_path.read_text().split('\n')

        requirements_file = example_root / 'requirements.txt'
        # will used what's installed in VIRTUAL_ENV already
        requirements_file.write_text('grizzly-loadtester\ngrizzly-loadtester-common\n')

        install_dependencies(e2e_fixture, example_root)

        index = feature_file_contents.index('  Scenario: dog facts api')
        # should go last in "Background"-section
        feature_file_contents.insert(index - 1, f'    Then start webserver on master port "{e2e_fixture.webserver_port}"')

        with feature_file_path.open('w') as fd:
            fd.truncate(0)
            fd.write('\n'.join(feature_file_contents))

        with NamedTemporaryFile(delete=False, suffix='.yaml', dir=f'{example_root}/environments') as env_conf_file:
            env_conf_file.write(yaml.dump(env_conf, Dumper=yaml.Dumper).encode())
            env_conf_file.flush()

            rc, output = e2e_fixture.execute(
                feature_file,
                env_conf_file.name.replace(f'{example_root.as_posix()}{pathsep}', ''),
                cwd=example_root,
                arguments=['-l', 'test_run.log'],
            )

            # problems with a locust DEBUG log message containing ERROR in the message on macos-latest
            if sys.platform == 'darwin':
                output = [line for line in output if 'ERROR' not in line and 'DEBUG' not in line]

            result = ''.join(output)

            validate_result(rc, result, example_root)
    except:
        if result is not None:
            with e2e_fixture.log_file.open('a+') as fd:
                fd.write(''.join(output))
        raise



"""Functionality for `grizzly-cli keyvault ...`."""

from __future__ import annotations

import logging
import re
from base64 import b64encode
from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any

import yaml
from azure.core.exceptions import ClientAuthenticationError, ResourceNotFoundError, ServiceRequestError

from grizzly_cli import register_parser
from grizzly_cli.argparse.bashcompletion import BashCompletionTypes
from grizzly_cli.utils import IndentDumper, chunker, flatten, logger, merge_dicts, unflatten
from grizzly_cli.utils.configuration import get_context_root, get_keyvault_client, load_configuration_file, load_configuration_keyvault

if TYPE_CHECKING:  # pragma: no cover
    from argparse import ArgumentParser as CoreArgumentParser
    from argparse import Namespace as Arguments

    from azure.keyvault.secrets import SecretClient

    from grizzly_cli.argparse import ArgumentSubParser

# disable azure.identity warning logs if authentication fails
azure_logger = logging.getLogger('azure')
azure_logger.setLevel(logging.ERROR)


KEYWORDS = ['password', 'secret', 'key', 'cert', 'token']
COMMON_FALSE_POSITIVES = ['RootManageSharedAccessKey']
KEYVAULT_MAX_SIZE = 25600 - 100  # 100 bytes buffer


@dataclass
class KeyvaultSecretHolder:
    name: str
    content_type: str | None
    value: str


@register_parser()
def create_parser(sub_parser: ArgumentSubParser) -> None:
    # grizzly-cli keyvault
    keyvault_parser = sub_parser.add_parser('keyvault', description=('grizzly keyvault integration'))

    keyvault_parser.add_argument(
        '-f',
        '--file',
        dest='env_file',
        type=BashCompletionTypes.File('*.yaml', '*.yml', missing_ok=True),
        required=True,
        help='path to the grizzly environment configuration file',
    )

    keyvault_parser.add_argument(
        '--vault-name',
        dest='keyvault',
        metavar='URL',
        type=str,
        help='keyvault url',
        default=None,
        required=False,
    )

    if keyvault_parser.prog != 'grizzly-cli keyvault':
        keyvault_parser.prog = 'grizzly-cli keyvault'

    sub_parser = keyvault_parser.add_subparsers(dest='subcommand')

    create_import_parser(sub_parser)
    create_export_parser(sub_parser)
    create_diff_parser(sub_parser)


def add_generic_arguments(parser: CoreArgumentParser) -> None:
    parser.add_argument(
        '--verbose',
        action='store_true',
        help='enable verbose output',
    )

    parser.add_argument(
        '--dry-run',
        action='store_true',
        help='do not write to keyvault',
    )

    parser.add_argument(
        '-k',
        '--key',
        action='append',
        dest='keys',
        type=str,
        required=False,
        help='filter on specified keys',
    )


def create_diff_parser(sub_parser: ArgumentSubParser) -> None:
    # grizzly-cli keyvault diff
    diff_parser = sub_parser.add_parser('diff', description='compare two environment configuration files')

    diff_parser.add_argument('orig_file', nargs=None, type=BashCompletionTypes.File('*.yaml', '*.yml'), help='file to compare -f/--file with')

    if diff_parser.prog != 'grizzly-cli keyvault diff':
        diff_parser.prog = 'grizzly-cli keyvault diff'


def create_import_parser(sub_parser: ArgumentSubParser) -> None:
    # grizzly-cli keyvault import
    import_parser = sub_parser.add_parser('import', description=('import grizzly environment configuration from keyvault secrets'))

    add_generic_arguments(import_parser)

    if import_parser.prog != 'grizzly-cli keyvault import':
        import_parser.prog = 'grizzly-cli keyvault import'


def create_export_parser(sub_parser: ArgumentSubParser) -> None:
    # grizzly-cli keyvault export
    export_parser = sub_parser.add_parser('export', description=('export grizzly environment configuration to keyvault secrets'))

    export_parser.add_argument(
        '-g',
        '--global',
        dest='global_configuration',
        metavar='KEY',
        action='append',
        default=[],
        type=str,
        required=False,
        help='environment configuration keys that are global',
    )

    add_generic_arguments(export_parser)

    if export_parser.prog != 'grizzly-cli keyvault export':
        export_parser.prog = 'grizzly-cli keyvault export'


def _keyvault_normalize(value: str) -> str:
    return re.sub(r'[^a-zA-Z0-9]', '-', value)


def encode_mq_certificate(root: Path, environment: str, key: str, base_cert_name: str) -> list[KeyvaultSecretHolder]:
    certificate_files = root.glob(f'{base_cert_name}.*')
    secrets: list[KeyvaultSecretHolder] = []

    references: list[str] = []

    for certificate_file in certificate_files:
        key_name = f'grizzly--{environment}--{_keyvault_normalize(certificate_file.name)}'
        references.append(key_name)
        secrets.extend(encode_file(key_name, certificate_file, no_conf=True))

    secrets.append(
        KeyvaultSecretHolder(
            name=key,
            content_type='files',
            value=','.join(references),
        ),
    )

    return secrets


def encode_file(key: str, file: str | Path, *, no_conf: bool = False) -> list[KeyvaultSecretHolder]:
    secrets: list[KeyvaultSecretHolder] = []
    if isinstance(file, str):
        file = Path(file)

    value = b64encode(file.read_bytes()).decode('utf-8')

    content_type_extra = ''

    if no_conf:
        content_type_extra = ',noconf'

    if len(value) > KEYVAULT_MAX_SIZE:
        references: list[str] = []
        chunks = chunker(value, KEYVAULT_MAX_SIZE)
        for index, chunk in enumerate(chunks):
            chunk_key = f'{key}--{index}'
            references.append(chunk_key)

            logger.debug(f'% creating chunk {index + 1} of {len(chunks)} in key {chunk_key}')

            secrets.append(
                KeyvaultSecretHolder(
                    name=chunk_key,
                    content_type=f'file:{file.name},chunk:{index},chunks:{len(chunks)}{content_type_extra}',
                    value=chunk,
                )
            )

        secrets.append(
            KeyvaultSecretHolder(
                name=key,
                content_type=f'files{content_type_extra}',
                value=','.join(references),
            )
        )
    else:
        secrets.append(
            KeyvaultSecretHolder(
                name=key,
                content_type=f'file:{file.name}{content_type_extra}',
                value=value,
            )
        )

    return secrets


def _should_export(key: str, secret: Any) -> bool:
    should: list[bool] = []

    if key == 'keyvault' and secret.endswith('.vault.azure.net'):
        return False

    for keyword in KEYWORDS:
        key_should = keyword in key.lower() and f'.{keyword}.' not in key
        secret_should = isinstance(secret, str) and keyword in secret.lower() and secret not in COMMON_FALSE_POSITIVES

        # if key should be exported, but not the secret and the secret is not a string... it shouldn't
        if key_should and not secret_should:
            key_should = isinstance(secret, str)

        should.append(key_should or secret_should)

    return any(should)


def _determine_environment(global_configuration: list[str], environment: str, key: str) -> str:
    return 'global' if any(key.startswith(keyword) for keyword in global_configuration) else environment


def _build_key_name(environment: str, key: str) -> str:
    return f'grizzly--{environment}--{_keyvault_normalize(key)}'


def _dict_to_yaml(file: Path, content: dict, *, indentation: Path | int) -> None:
    file.write_text('')  # make sure file is empty

    with file.open('w') as fd:
        yaml.dump(content, fd, Dumper=IndentDumper.use_indentation(indentation), default_flow_style=False, sort_keys=False, allow_unicode=True)


def _extract_metadata(env_file: str) -> tuple[str, str | None, dict]:
    file = Path(env_file)
    configuration = load_configuration_file(file).get('configuration', {})

    keyvault = configuration.get('keyvault', None)

    configuration = flatten(configuration)

    logger.debug('flatten configuration: %r', configuration)

    return (configuration.get('env', None) or file.stem, keyvault, configuration)


def diff(left_file_name: str, right_file_name: str) -> int:
    left_config_file = Path(left_file_name)
    right_config_file = Path(right_file_name)

    if not left_config_file.exists():
        message = f'environment configuration file {left_config_file.as_posix()} does not exist'
        raise ValueError(message)

    if not right_config_file.exists():
        message = f'environment configuration file {right_config_file.as_posix()} does not exist'
        raise ValueError(message)

    left_config = flatten(load_configuration_file(right_config_file)['configuration'])
    right_config = flatten(load_configuration_file(left_config_file)['configuration'])

    logger.info(f'- {right_config_file.as_posix()}')
    logger.info(f'+ {left_config_file.as_posix()}')
    logger.info('')

    diffed_keys: set[str] = set()

    for key, value in left_config.items():
        if key not in right_config or right_config.get(key, None) != value:
            logger.error(f'- {key}: {value} != {right_config.get(key, None)}')
            diffed_keys.add(key)

    for key, value in right_config.items():
        if key not in left_config or (left_config.get(key, None) != value and key not in diffed_keys):
            logger.error(f'+ {key}: {left_config.get(key, None)} != {value}')

    return 0


def keyvault_import(client: SecretClient, environment: str, args: Arguments, root: Path, configuration: dict) -> int:
    # unflatten existing configuration
    configuration_unflatten: dict = {}

    original_key_count = len(configuration)

    for conf_key, conf_value in configuration.items():
        configuration_branch = unflatten(conf_key, conf_value)
        configuration_unflatten = merge_dicts(configuration_branch, configuration_unflatten)

    if len(configuration_unflatten) < 1:
        logger.error('environment file %s did not contain any configuration', args.env_file)
        return 1

    configuration = configuration_unflatten

    logger.debug('exploded configuration: %r', configuration)

    keyvault_configuration, imported_secrets = load_configuration_keyvault(client, environment, root, filter_keys=args.keys)

    configuration = merge_dicts(keyvault_configuration, configuration)

    logger.debug('merged configuration: %r', configuration)

    environment_file = Path(args.env_file)

    if not args.dry_run:  # do not rewrite environment file on dry-run
        flatten_configuration = flatten(configuration)
        written_keys = len(flatten_configuration) - original_key_count
        unsafe_environment_file = environment_file.rename(environment_file.with_suffix(f'.unsafe{environment_file.suffix}'))
        _dict_to_yaml(environment_file, {'configuration': configuration}, indentation=environment_file)
    else:
        written_keys = 0
        unsafe_environment_file = environment_file.with_suffix(f'.unsafe{environment_file.suffix}')

    logger.info(
        '\nimported %d secrets from %s and wrote %d new keys to %s, saved original in %s',
        imported_secrets,
        client.vault_url,
        written_keys,
        environment_file.as_posix(),
        unsafe_environment_file.as_posix(),
    )

    return 0


def upsert_secrets(client: SecretClient, secrets: list[KeyvaultSecretHolder], *, dry_run: bool) -> int:
    created_secrets_count = 0

    for secret in secrets:
        # check if secret already exists
        try:
            current_value = client.get_secret(secret.name)
            if current_value.value != secret.value or current_value.properties.content_type != secret.content_type:
                raise ResourceNotFoundError

            # it exists with the same value, skip
            continue
        except ResourceNotFoundError:  # great, import it into the keyvault
            pass

        logger.debug(f'% keyvault secret {secret.name} with content type {secret.content_type}')

        if not dry_run:
            client.set_secret(secret.name, secret.value, content_type=secret.content_type)

        created_secrets_count += 1

    return created_secrets_count


def prepare_secrets(
    client: SecretClient,
    environment: str,
    root: Path,
    configuration: dict,
    *,
    filter_keys: list[str] | None,
    global_configuration: list[str],
) -> tuple[dict, list[KeyvaultSecretHolder]]:
    secrets: list[KeyvaultSecretHolder] = []
    safe_configuration: dict = {}

    for key, secret in configuration.items():
        if not _should_export(key, secret):
            safe_configuration.update({key: secret})
            continue

        if filter_keys is not None and key not in filter_keys:
            continue

        key_environment = _determine_environment(global_configuration, environment, key)
        key_name = _build_key_name(key_environment, key)

        if secret.startswith('cert:'):
            if '#' in secret:
                secret, content_type = secret.split('#', 1)  # noqa: PLW2901
            else:
                content_type = None

            if 'pass:' in secret:
                _, password_ref = secret.split(',', 1)
                _, password_key = password_ref.split(':', 1)
                try:
                    client.get_secret(password_key)
                except ResourceNotFoundError:
                    message = f'key {password_key} referenced in value for {key} does not exist'
                    raise ValueError(message) from None

            secrets.append(
                KeyvaultSecretHolder(
                    name=key_name,
                    content_type=content_type,
                    value=secret,
                )
            )
        elif 'file' in key:
            if 'mq' in key:
                secrets.extend(encode_mq_certificate(root, key_environment, key_name, secret))
            else:
                secrets.extend(encode_file(key_name, secret))
        else:
            secrets.append(
                KeyvaultSecretHolder(
                    name=key_name,
                    content_type=None,
                    value=secret,
                )
            )

    return safe_configuration, secrets


def keyvault_export(client: SecretClient, environment: str, args: Arguments, root: Path, configuration: dict) -> int:
    """From grizzly to keyvault.

    If the specified secret starts with `cert:`, it indicates that it should reference a keyvault certificate (the name). This value
    can have metadata (secret content type) appended after the actual value the `#` separator. If the output certificate should be
    password protected, `pass:` should reference a keyvault secret which contains the password.

    `cert:<keyvault certificate name>[,pass:<keyvault secret name for password>][#format:[mqm|pem-public|pem-private]]`

    Supported certificate output formats:
    - `mqm`, MQ CMS keystore
    - `pem-public`, public certificate in PEM format
    - `pem-private`, private key in PEM format

    If the configuration key contains `file` in the path, the configuration value will be base64 encoded and, optionally, chunked into
    keyvault secrets. If the path also contains `mq`, all MQ keystore/CMS files will be encoded, chunked and then references to the
    actual configuration key.
    """
    environment_file = Path(args.env_file)

    safe_configuration, secrets = prepare_secrets(client, environment, root, configuration, filter_keys=args.keys, global_configuration=args.global_configuration)

    created_secrets_count = upsert_secrets(client, secrets, dry_run=args.dry_run)
    already_exists = len(secrets) - created_secrets_count

    if not args.dry_run:
        unsafe_environment_file = environment_file.rename(environment_file.with_suffix(f'.unsafe{environment_file.suffix}'))
        safe_yaml_configuration: dict = {}

        for key, value in safe_configuration.items():
            safe_yaml_configuration = merge_dicts(safe_yaml_configuration, unflatten(key, value))

        if safe_yaml_configuration.get('keyvault', None) is None:
            safe_yaml_configuration.update({'keyvault': client.vault_url})

        safe_yaml_configuration = {'configuration': safe_yaml_configuration}

        # remove contents in safe file, before writing the safe configuration
        _dict_to_yaml(environment_file, safe_yaml_configuration, indentation=unsafe_environment_file)
    else:
        unsafe_environment_file = environment_file.with_suffix(f'.unsafe{environment_file.suffix}')

    logger.info(
        f'created {created_secrets_count} ({already_exists} already existed) secrets in keyvault {client.vault_url} '
        f'and saved the safe environment configuration in {environment_file.as_posix()}',
    )
    logger.warning(f'! the unsafe environment configuration is still present in {unsafe_environment_file.as_posix()}')

    return 0


def keyvault(args: Arguments) -> int:
    grizzly_context_root = get_context_root()

    with suppress(AttributeError):
        if args.verbose or args.dry_run:
            logger.setLevel(logging.DEBUG)

    args_keyvault = f'https://{args.keyvault}.vault.azure.net' if args.keyvault is not None else None

    env_file = Path(args.env_file)
    if args.subcommand == 'import' and not env_file.exists():
        if args.keyvault is None:
            message = f'--vault-name not specified and environment configuration file {args.env_file} does not exist'
            raise ValueError(message)

        _dict_to_yaml(env_file, {'configuration': {'keyvault': args_keyvault}}, indentation=2)

    environment, keyvault, configuration = _extract_metadata(args.env_file)

    if keyvault is None and args.keyvault is None:
        message = 'keyvault not specified in environment configuration file and no keyvault specified on the command line'
        raise ValueError(message)

    keyvault = keyvault or args_keyvault

    if keyvault is None:
        message = 'keyvault not specified, please specify a keyvault'
        raise ValueError(message)

    client = get_keyvault_client(keyvault)

    try:
        if args.subcommand == 'import':  # from keyvault
            return keyvault_import(client, environment, args, grizzly_context_root, configuration)

        if args.subcommand == 'export':  # to keyvault
            return keyvault_export(client, environment, args, grizzly_context_root, configuration)

        if args.subcommand == 'diff':
            return diff(args.env_file, args.orig_file)

        message = f'unknown subcommand {args.subcommand}'
        raise ValueError(message)
    except ClientAuthenticationError:
        logger.error('authentication failed, if you are running from a resource which does not have a managed identity then you must run `az login` first.')
        return 1
    except ServiceRequestError:
        logger.error(f'{client.vault_url} does not resolve to an azure keyvault')
        return 1
    except ValueError:
        raise



"""Functionality for `grizzly dist ...`."""

from __future__ import annotations

import argparse
import os
import subprocess
import sys
from argparse import Namespace as Arguments
from getpass import getuser
from io import StringIO
from json import loads as jsonloads
from pathlib import Path
from shutil import get_terminal_size
from socket import gethostname
from tempfile import NamedTemporaryFile
from typing import IO, TYPE_CHECKING

from grizzly_cli import EXECUTION_CONTEXT, MOUNT_CONTEXT, PROJECT_NAME, STATIC_CONTEXT, register_parser
from grizzly_cli.distributed.build import build as do_build
from grizzly_cli.distributed.build import create_parser as build_create_parser
from grizzly_cli.distributed.clean import clean as do_clean
from grizzly_cli.distributed.clean import create_parser as clean_create_parser
from grizzly_cli.run import create_parser as run_create_parser
from grizzly_cli.run import run
from grizzly_cli.utils import (
    get_default_mtu,
    list_images,
    run_command,
)

if TYPE_CHECKING:  # pragma: no cover
    from grizzly_cli.argparse import ArgumentSubParser


@register_parser(order=3)
def create_parser(sub_parser: ArgumentSubParser) -> None:
    dist_parser = sub_parser.add_parser('dist', description='commands for running grizzly i distributed mode.')

    dist_parser.add_argument(
        '--workers',
        type=int,
        required=False,
        default=1,
        help='how many instances of the `workers` container that should be created',
    )
    dist_parser.add_argument(
        '--container-system',
        type=str,
        choices=['podman', 'docker', None],
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    dist_parser.add_argument(
        '--id',
        type=str,
        required=False,
        default=None,
        help='unique identifier suffixed to compose project, should be used when the same user needs to run more than one instance of `grizzly-cli`',
    )
    dist_parser.add_argument(
        '--limit-nofile',
        type=int,
        required=False,
        default=10001,
        help='set system limit "number of open files"',
    )
    dist_parser.add_argument(
        '--health-retries',
        type=int,
        required=False,
        default=3,
        help='set number of retries for health check of master container',
    )
    dist_parser.add_argument(
        '--health-timeout',
        type=int,
        required=False,
        default=3,
        help='set timeout in seconds for health check of master container',
    )
    dist_parser.add_argument(
        '--health-interval',
        type=int,
        required=False,
        default=5,
        help='set interval in seconds between health checks of master container',
    )
    dist_parser.add_argument(
        '--registry',
        type=str,
        default=None,
        required=False,
        help='push built image to this registry, if the registry has authentication you need to login first',
    )
    dist_parser.add_argument(
        '--tty',
        action='store_true',
        default=False,
        required=False,
        help='start containers with a TTY enabled',
    )
    dist_parser.add_argument(
        '--wait-for-worker',
        type=str,
        default=None,
        required=False,
        help=('sets enviroment variable LOCUST_WAIT_FOR_WORKERS_REPORT_AFTER_RAMP_UP, which tells master to wait this amount of time for worker report'),
    )

    dist_parser.add_argument(
        '--project-name',
        type=str,
        default=None,
        help='override project name, which otherwise would be the name of the directory where command is executed in',
    )

    group_build = dist_parser.add_mutually_exclusive_group()
    group_build.add_argument(
        '--force-build',
        action='store_true',
        required=False,
        help='force rebuild the grizzly projects container image (no cache)',
    )
    group_build.add_argument(
        '--build',
        action='store_true',
        required=False,
        help='rebuild the grizzly projects container images (with cache)',
    )
    group_build.add_argument(
        '--validate-config',
        action='store_true',
        required=False,
        help='validate and print compose project file',
    )

    if dist_parser.prog != 'grizzly-cli dist':  # pragma: no cover
        dist_parser.prog = 'grizzly-cli dist'

    sub_parser = dist_parser.add_subparsers(dest='subcommand')

    build_create_parser(sub_parser)
    clean_create_parser(sub_parser)
    run_create_parser(sub_parser, parent='dist')


def distributed(args: Arguments) -> int:
    if args.subcommand == 'run':
        return run(args, distributed_run)

    if args.subcommand == 'build':
        return do_build(args)
    if args.subcommand == 'clean':
        return do_clean(args)

    message = f'unknown subcommand {args.subcommand}'
    raise ValueError(message)


def update_os_environ(args: Arguments, run_arguments: dict[str, list[str]], project_name: str, tag: str) -> None:
    if args.file is not None:
        os.environ['GRIZZLY_RUN_FILE'] = args.file

    if args.wait_for_worker is not None:
        os.environ['LOCUST_WAIT_FOR_WORKERS_REPORT_AFTER_RAMP_UP'] = f'{args.wait_for_worker}'

    mtu = get_default_mtu(args)

    if mtu is None and os.environ.get('GRIZZLY_MTU', None) is None:
        print('!! unable to determine MTU, try manually setting GRIZZLY_MTU environment variable if anything other than 1500 is needed')
        mtu = '1500'

    columns, lines = get_terminal_size()

    # set environment variables needed by compose files, when * compose executes
    os.environ.update(
        {
            'GRIZZLY_MTU': str(mtu),
            'GRIZZLY_EXECUTION_CONTEXT': EXECUTION_CONTEXT,
            'GRIZZLY_STATIC_CONTEXT': STATIC_CONTEXT,
            'GRIZZLY_MOUNT_CONTEXT': MOUNT_CONTEXT,
            'GRIZZLY_PROJECT_NAME': project_name,
            'GRIZZLY_USER_TAG': tag,
            'GRIZZLY_EXPECTED_WORKERS': str(args.workers),
            'GRIZZLY_LIMIT_NOFILE': str(args.limit_nofile),
            'GRIZZLY_HEALTH_CHECK_RETRIES': str(args.health_retries),
            'GRIZZLY_HEALTH_CHECK_INTERVAL': str(args.health_interval),
            'GRIZZLY_HEALTH_CHECK_TIMEOUT': str(args.health_timeout),
            'GRIZZLY_IMAGE_REGISTRY': getattr(args, 'registry', None) or '',
            'GRIZZLY_CONTAINER_TTY': repr(args.tty).lower(),
            'COLUMNS': str(columns),
            'LINES': str(lines),
        }
    )

    grizzly_mount_context_path = ''

    if EXECUTION_CONTEXT != MOUNT_CONTEXT:
        hostname = gethostname()
        output = subprocess.check_output(
            [args.container_system, 'container', 'inspect', '-f', '{{ json .Mounts }}', hostname],
            encoding='utf-8',
        )
        container_mounts = jsonloads(output)
        for container_mount in container_mounts:
            if container_mount['Source'] != MOUNT_CONTEXT:
                continue

            grizzly_mount_context_path = EXECUTION_CONTEXT.replace(container_mount['Destination'], '')[1:]
            break

    os.environ['GRIZZLY_MOUNT_PATH'] = grizzly_mount_context_path

    if len(run_arguments.get('master', [])) > 0:
        os.environ['GRIZZLY_MASTER_RUN_ARGS'] = ' '.join(run_arguments['master'])

    if len(run_arguments.get('worker', [])) > 0:
        os.environ['GRIZZLY_WORKER_RUN_ARGS'] = ' '.join(run_arguments['worker'])

    if len(run_arguments.get('common', [])) > 0:
        os.environ['GRIZZLY_COMMON_RUN_ARGS'] = ' '.join(run_arguments['common'])


def write_env_file(fd: IO[bytes], environ: dict, args: Arguments) -> None:
    if len(environ) > 0:
        for key, value in environ.items():
            transformed_value = value.replace(EXECUTION_CONTEXT, MOUNT_CONTEXT).replace(MOUNT_CONTEXT, '/srv/grizzly') if key == 'GRIZZLY_CONFIGURATION_FILE' else value

            fd.write(f'{key}={transformed_value}\n'.encode())

    fd.write(f'COLUMNS={os.environ["COLUMNS"]}\n'.encode())
    fd.write(f'LINES={os.environ["LINES"]}\n'.encode())
    fd.write(f'GRIZZLY_CONTAINER_TTY={os.environ["GRIZZLY_CONTAINER_TTY"]}\n'.encode())
    fd.write(b'GRIZZLY_RUN_MODE="distributed"\n')

    if args.wait_for_worker is not None:
        fd.write(f'LOCUST_WAIT_FOR_WORKERS_REPORT_AFTER_RAMP_UP="{args.wait_for_worker}"'.encode())

    fd.flush()

    os.environ['GRIZZLY_ENVIRONMENT_FILE'] = fd.name


def should_validate_config(args: Arguments, compose_args: list[str]) -> int:
    validate_config = getattr(args, 'validate_config', False)

    compose_command = [
        args.container_system,
        'compose',
        *compose_args,
        'config',
    ]

    result = run_command(compose_command, silent=not validate_config)

    if validate_config or result.return_code != 0:
        if result.return_code != 0 and not validate_config:
            print('!! something in the compose project is not valid, check with:')
            argv = sys.argv[:]
            argv.insert(argv.index('dist') + 1, '--validate-config')
            print(f'grizzly-cli {" ".join(argv[1:])}')

        return result.return_code

    return 0


def should_build_image(args: Arguments, project_name: str, tag: str) -> int:
    images = list_images(args)

    if images.get(project_name, {}).get(tag, None) is None or args.force_build or args.build:
        rc = do_build(args)
        if rc != 0:
            print(f'!! failed to build {project_name}, rc={rc}')
            return rc

    return 0


def distributed_run(args: Arguments, environ: dict, run_arguments: dict[str, list[str]]) -> int:
    suffix = '' if args.id is None else f'-{args.id}'
    tag = getuser()

    project_name = PROJECT_NAME if args.project_name is None else args.project_name

    # default locust project
    compose_args: list[str] = [
        '-p',
        f'{project_name}{suffix}-{tag}',
        '-f',
        f'{STATIC_CONTEXT}/compose.yaml',
    ]

    update_os_environ(args, run_arguments, project_name, tag)

    name_template = '{project}{suffix}-{tag}-{node}-{index}'

    with NamedTemporaryFile() as fd:  # file will be deleted when container exists
        write_env_file(fd, environ, args)

        rc = should_validate_config(args, compose_args)
        if rc != 0 or getattr(args, 'validate_config', False):
            return rc

        rc = should_build_image(args, project_name, tag)
        if rc != 0:
            return rc

        compose_scale_argument = ['--scale', f'worker={args.workers}']

        # bring up containers
        compose_command = [
            args.container_system,
            'compose',
            *compose_args,
            'up',
            *compose_scale_argument,
            '--remove-orphans',
        ]

        result = run_command(compose_command, verbose=args.verbose)

        try:
            output = subprocess.check_output(
                [
                    args.container_system,
                    'inspect',
                    '-f',
                    '{{ .State.ExitCode }}',
                    name_template.format(
                        project=project_name,
                        suffix=suffix,
                        tag=tag,
                        node='master',
                        index=1,
                    ),
                ],
                encoding='utf-8',
            )
            result.return_code = int(output.strip())
        except:
            result.return_code = 1

        # stop containers
        compose_command = [
            args.container_system,
            'compose',
            *compose_args,
            'stop',
        ]

        run_command(compose_command)

        if result.return_code != 0:
            if result.abort_timestamp is not None:
                master_node_name = name_template.format(
                    project=project_name,
                    suffix=suffix,
                    tag=tag,
                    node='master',
                    index=1,
                )

                since_timestamp = result.abort_timestamp.strftime('%Y-%m-%dT%H:%M:%SZ')
                command = [args.container_system, 'container', 'logs', '--since', since_timestamp, master_node_name]

                missed_output = subprocess.check_output(
                    command,
                    encoding='utf-8',
                    shell=False,
                    universal_newlines=True,
                    stderr=subprocess.STDOUT,
                ).split('\n')

                log_file = Path(args.log_file).open('a+') if args.log_file is not None else StringIO()  # noqa: SIM115

                try:
                    for line in missed_output:
                        formatted_line = f'{master_node_name}  | {line}'
                        print(formatted_line)
                        log_file.write(f'{formatted_line}\n')
                except:  # noqa: S110
                    pass
                finally:
                    log_file.close()

            print('\n!! something went wrong, check full container logs with:')
            template = '{container_system} container logs {name_template}'
            print(
                template.format(
                    container_system=args.container_system,
                    name_template=name_template.format(
                        project=project_name,
                        suffix=suffix,
                        tag=tag,
                        node='master',
                        index=1,
                    ),
                )
            )

            for worker in range(1, args.workers + 1):
                print(
                    template.format(
                        container_system=args.container_system,
                        name_template=name_template.format(
                            project=project_name,
                            suffix=suffix,
                            tag=tag,
                            node='worker',
                            index=worker,
                        ),
                    )
                )

        return result.return_code



"""Tests for grizzly_cli.distributed."""

from __future__ import annotations

import json
import sys
from argparse import ArgumentParser, Namespace
from contextlib import suppress
from datetime import datetime, timezone
from os import environ
from tempfile import gettempdir
from typing import TYPE_CHECKING

import pytest
from grizzly_cli.distributed import create_parser, distributed, distributed_run
from grizzly_cli.utils import RunCommandResult, rm_rf

if TYPE_CHECKING:  # pragma: no cover
    from _pytest.capture import CaptureFixture
    from _pytest.tmpdir import TempPathFactory
    from pytest_mock import MockerFixture


def test_distributed(mocker: MockerFixture) -> None:
    run_mocked = mocker.patch('grizzly_cli.distributed.run', return_value=0)
    build_mocked = mocker.patch('grizzly_cli.distributed.do_build', return_value=5)
    clean_mocked = mocker.patch('grizzly_cli.distributed.do_clean', return_value=10)

    arguments = Namespace(subcommand='run')
    assert distributed(arguments) == 0

    assert run_mocked.call_count == 1
    args, _ = run_mocked.call_args_list[0]
    assert args[0] is arguments
    assert args[1] is distributed_run

    assert build_mocked.call_count == 0
    assert clean_mocked.call_count == 0

    arguments = Namespace(subcommand='build')
    assert distributed(arguments) == 5

    assert run_mocked.call_count == 1
    assert build_mocked.call_count == 1
    assert clean_mocked.call_count == 0
    args, _ = build_mocked.call_args_list[0]
    assert args[0] is arguments

    arguments = Namespace(subcommand='clean')
    assert distributed(arguments) == 10

    assert run_mocked.call_count == 1
    assert build_mocked.call_count == 1
    assert clean_mocked.call_count == 1
    args, _ = clean_mocked.call_args_list[0]
    assert args[0] is arguments

    arguments = Namespace(subcommand='foo')
    with pytest.raises(ValueError, match='unknown subcommand foo'):
        distributed(arguments)


def test_distributed_run(capsys: CaptureFixture, mocker: MockerFixture, tmp_path_factory: TempPathFactory) -> None:  # noqa: PLR0915
    test_context = tmp_path_factory.mktemp('test_context')
    (test_context / 'test.feature').write_text('Feature:')

    mocker.patch('grizzly_cli.distributed.getuser', return_value='test-user')
    get_default_mtu_mock = mocker.patch('grizzly_cli.distributed.get_default_mtu', return_value=None)
    do_build_mock = mocker.patch('grizzly_cli.distributed.do_build', return_value=None)
    list_images_mock = mocker.patch('grizzly_cli.distributed.list_images', return_value=None)

    import grizzly_cli.distributed

    mocker.patch.object(grizzly_cli.distributed, 'EXECUTION_CONTEXT', '/srv/grizzly/execution-context')
    mocker.patch.object(grizzly_cli.distributed, 'STATIC_CONTEXT', '/srv/grizzly/static-context')
    mocker.patch.object(grizzly_cli.distributed, 'MOUNT_CONTEXT', '/srv/grizzly/mount-context')
    mocker.patch.object(grizzly_cli.distributed, 'PROJECT_NAME', 'grizzly-cli-test-project')

    run_command_result = RunCommandResult(return_code=1)
    run_command_result.abort_timestamp = datetime.now(tz=timezone.utc)

    run_command_mock = mocker.patch('grizzly_cli.distributed.run_command', return_value=None)
    check_output_mock = mocker.patch('grizzly_cli.distributed.subprocess.check_output', return_value=None)

    parser = ArgumentParser()

    sub_parsers = parser.add_subparsers(dest='test')

    create_parser(sub_parsers)

    try:
        run_command_mock.return_value = RunCommandResult(return_code=111)
        check_output_mock.return_value = '{}'
        get_default_mtu_mock.return_value = '1500'
        sys.argv = ['grizzly-cli', 'dist', '--workers', '3', '--tty', 'run', f'{test_context}/test.feature']
        arguments = parser.parse_args()
        setattr(arguments, 'container_system', 'docker')  # noqa: B010
        setattr(arguments, 'file', ' '.join(arguments.file))  # noqa: B010

        # this is set in the devcontainer
        for key in environ:
            if key.startswith('GRIZZLY_'):
                del environ[key]

        assert distributed_run(arguments, {}, {}) == 111
        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == (
            f'!! something in the compose project is not valid, check with:\ngrizzly-cli dist --validate-config --workers 3 --tty run {test_context}/test.feature\n'
        )

        with suppress(KeyError):
            del environ['GRIZZLY_MTU']

        run_command_mock.return_value = RunCommandResult(return_code=0)
        do_build_mock.return_value = 255
        check_output_mock.return_value = '{}'
        get_default_mtu_mock.return_value = None
        list_images_mock.return_value = {}

        assert distributed_run(arguments, {}, {}) == 255
        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == (
            '!! unable to determine MTU, try manually setting GRIZZLY_MTU environment variable if anything other than 1500 is needed\n'
            '!! failed to build grizzly-cli-test-project, rc=255\n'
        )
        assert environ.get('GRIZZLY_MTU', None) == '1500'
        assert environ.get('GRIZZLY_EXECUTION_CONTEXT', None) == '/srv/grizzly/execution-context'
        assert environ.get('GRIZZLY_STATIC_CONTEXT', None) == '/srv/grizzly/static-context'
        assert environ.get('GRIZZLY_MOUNT_CONTEXT', None) == '/srv/grizzly/mount-context'
        assert environ.get('GRIZZLY_PROJECT_NAME', None) == 'grizzly-cli-test-project'
        assert environ.get('GRIZZLY_USER_TAG', None) == 'test-user'
        assert environ.get('GRIZZLY_EXPECTED_WORKERS', None) == '3'
        assert environ.get('GRIZZLY_MASTER_RUN_ARGS', None) is None
        assert environ.get('GRIZZLY_WORKER_RUN_ARGS', None) is None
        assert environ.get('GRIZZLY_COMMON_RUN_ARGS', None) is None
        assert environ.get('GRIZZLY_IMAGE_REGISTRY', None) == ''
        assert environ.get('GRIZZLY_ENVIRONMENT_FILE', '').startswith(gettempdir())
        assert environ.get('GRIZZLY_LIMIT_NOFILE', None) == '10001'
        assert environ.get('GRIZZLY_HEALTH_CHECK_INTERVAL', None) == '5'
        assert environ.get('GRIZZLY_HEALTH_CHECK_TIMEOUT', None) == '3'
        assert environ.get('GRIZZLY_HEALTH_CHECK_RETRIES', None) == '3'
        assert environ.get('GRIZZLY_CONTAINER_TTY', None) == 'true'
        assert environ.get('LOCUST_WAIT_FOR_WORKERS_REPORT_AFTER_RAMP_UP', None) is None
        assert environ.get('GRIZZLY_MOUNT_PATH', None) == ''

        # this is set in the devcontainer
        for key in environ:
            if key.startswith(('GRIZZLY_', 'LOCUST_')):
                del environ[key]

        arguments = parser.parse_args(
            [
                'dist',
                '--workers',
                '3',
                '--build',
                '--limit-nofile',
                '133700',
                '--health-interval',
                '10',
                '--health-timeout',
                '8',
                '--health-retries',
                '30',
                '--registry',
                'registry.example.com/biometria-se',
                '--wait-for-worker',
                '10000',
                '--project-name',
                'foobar',
                'run',
                f'{test_context}/test.feature',
            ]
        )
        setattr(arguments, 'container_system', 'docker')  # noqa: B010
        setattr(arguments, 'file', ' '.join(arguments.file))  # noqa: B010

        # docker-compose v2
        rcr = RunCommandResult(return_code=1)
        rcr.abort_timestamp = datetime.now(tz=timezone.utc)
        run_command_mock.return_value = None
        run_command_mock.side_effect = [RunCommandResult(return_code=0), rcr, RunCommandResult(return_code=0)]
        do_build_mock.return_value = 0
        check_output_mock.return_value = None
        check_output_mock.side_effect = ['{}', '{}', '<!-- here is the missing logs -->']
        get_default_mtu_mock.return_value = '1400'
        list_images_mock.return_value = {'grizzly-cli-test-project': {'test-user': {}}}

        assert (
            distributed_run(
                arguments,
                {
                    'GRIZZLY_CONFIGURATION_FILE': '/srv/grizzly/execution-context/configuration.yaml',
                    'GRIZZLY_TEST_VAR': 'True',
                },
                {
                    'master': ['--foo', 'bar', '--master'],
                    'worker': ['--bar', 'foo', '--worker'],
                    'common': ['--common', 'true', '-Dcsv-prefix=asdf'],
                },
            )
            == 1
        )
        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == (
            'foobar-test-user-master-1  | <!-- here is the missing logs -->\n'
            '\n!! something went wrong, check full container logs with:\n'
            'docker container logs foobar-test-user-master-1\n'
            'docker container logs foobar-test-user-worker-1\n'
            'docker container logs foobar-test-user-worker-2\n'
            'docker container logs foobar-test-user-worker-3\n'
        )

        assert run_command_mock.call_count == 5
        args, _ = run_command_mock.call_args_list[-3]
        assert args[0] == [
            'docker',
            'compose',
            '-p',
            'foobar-test-user',
            '-f',
            '/srv/grizzly/static-context/compose.yaml',
            'config',
        ]
        args, _ = run_command_mock.call_args_list[-2]
        assert args[0] == [
            'docker',
            'compose',
            '-p',
            'foobar-test-user',
            '-f',
            '/srv/grizzly/static-context/compose.yaml',
            'up',
            '--scale',
            'worker=3',
            '--remove-orphans',
        ]
        args, _ = run_command_mock.call_args_list[-1]
        assert args[0] == [
            'docker',
            'compose',
            '-p',
            'foobar-test-user',
            '-f',
            '/srv/grizzly/static-context/compose.yaml',
            'stop',
        ]

        assert environ.get('GRIZZLY_RUN_FILE', None) == f'{test_context}/test.feature'
        assert environ.get('GRIZZLY_MTU', None) == '1400'
        assert environ.get('GRIZZLY_EXECUTION_CONTEXT', None) == '/srv/grizzly/execution-context'
        assert environ.get('GRIZZLY_STATIC_CONTEXT', None) == '/srv/grizzly/static-context'
        assert environ.get('GRIZZLY_MOUNT_CONTEXT', None) == '/srv/grizzly/mount-context'
        assert environ.get('GRIZZLY_PROJECT_NAME', None) == 'foobar'
        assert environ.get('GRIZZLY_USER_TAG', None) == 'test-user'
        assert environ.get('GRIZZLY_EXPECTED_WORKERS', None) == '3'
        assert environ.get('GRIZZLY_MASTER_RUN_ARGS', None) == '--foo bar --master'
        assert environ.get('GRIZZLY_WORKER_RUN_ARGS', None) == '--bar foo --worker'
        assert environ.get('GRIZZLY_COMMON_RUN_ARGS', None) == '--common true -Dcsv-prefix=asdf'
        assert environ.get('GRIZZLY_ENVIRONMENT_FILE', '').startswith(gettempdir())
        assert environ.get('GRIZZLY_LIMIT_NOFILE', None) == '133700'
        assert environ.get('GRIZZLY_HEALTH_CHECK_INTERVAL', None) == '10'
        assert environ.get('GRIZZLY_HEALTH_CHECK_TIMEOUT', None) == '8'
        assert environ.get('GRIZZLY_HEALTH_CHECK_RETRIES', None) == '30'
        assert environ.get('GRIZZLY_IMAGE_REGISTRY', None) == 'registry.example.com/biometria-se'
        assert environ.get('GRIZZLY_CONTAINER_TTY', None) == 'false'
        assert environ.get('LOCUST_WAIT_FOR_WORKERS_REPORT_AFTER_RAMP_UP', None) == '10000'
        assert environ.get('GRIZZLY_MOUNT_PATH', None) == ''

        arguments.project_name = None

        # this is set in the devcontainer
        for key in environ:
            if key.startswith(('GRIZZLY_', 'LOCUST_')):
                del environ[key]

        arguments = parser.parse_args(
            [
                'dist',
                '--workers',
                '1',
                '--id',
                'suffix',
                '--validate-config',
                '--limit-nofile',
                '20000',
                '--health-interval',
                '10',
                '--health-timeout',
                '8',
                '--health-retries',
                '30',
                '--wait-for-worker',
                '1.25 * WORKER_REPORT_INTERVAL',
                'run',
                f'{test_context}/test.feature',
            ]
        )
        setattr(arguments, 'container_system', 'docker')  # noqa: B010
        setattr(arguments, 'file', ' '.join(arguments.file))  # noqa: B010

        run_command_mock.return_value = None
        run_command_mock.side_effect = [RunCommandResult(return_code=13)]
        do_build_mock.return_value = 0
        check_output_mock.return_value = None
        check_output_mock.side_effect = [json.dumps([{'Source': '/srv/grizzly/mount-context', 'Destination': '/srv/grizzly'}]), '13']
        get_default_mtu_mock.return_value = '1800'
        list_images_mock.return_value = {'grizzly-cli-test-project': {'test-user': {}}}

        assert (
            distributed_run(
                arguments,
                {
                    'GRIZZLY_CONFIGURATION_FILE': '/srv/grizzly/execution-context/configuration.yaml',
                    'GRIZZLY_TEST_VAR': 'True',
                },
                {
                    'master': ['--foo', 'bar', '--master'],
                    'worker': ['--bar', 'foo', '--worker'],
                    'common': ['--common', 'true'],
                },
            )
            == 13
        )
        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == ''

        assert run_command_mock.call_count == 6
        args, _ = run_command_mock.call_args_list[-1]
        assert args[0] == [
            'docker',
            'compose',
            '-p',
            'grizzly-cli-test-project-suffix-test-user',
            '-f',
            '/srv/grizzly/static-context/compose.yaml',
            'config',
        ]

        assert environ.get('GRIZZLY_RUN_FILE', None) == f'{test_context}/test.feature'
        assert environ.get('GRIZZLY_MTU', None) == '1800'
        assert environ.get('GRIZZLY_EXECUTION_CONTEXT', None) == '/srv/grizzly/execution-context'
        assert environ.get('GRIZZLY_STATIC_CONTEXT', None) == '/srv/grizzly/static-context'
        assert environ.get('GRIZZLY_MOUNT_CONTEXT', None) == '/srv/grizzly/mount-context'
        assert environ.get('GRIZZLY_PROJECT_NAME', None) == 'grizzly-cli-test-project'
        assert environ.get('GRIZZLY_USER_TAG', None) == 'test-user'
        assert environ.get('GRIZZLY_EXPECTED_WORKERS', None) == '1'
        assert environ.get('GRIZZLY_MASTER_RUN_ARGS', None) == '--foo bar --master'
        assert environ.get('GRIZZLY_WORKER_RUN_ARGS', None) == '--bar foo --worker'
        assert environ.get('GRIZZLY_COMMON_RUN_ARGS', None) == '--common true'
        assert environ.get('GRIZZLY_ENVIRONMENT_FILE', '').startswith(gettempdir())
        assert environ.get('GRIZZLY_LIMIT_NOFILE', None) == '20000'
        assert environ.get('GRIZZLY_HEALTH_CHECK_INTERVAL', None) == '10'
        assert environ.get('GRIZZLY_HEALTH_CHECK_TIMEOUT', None) == '8'
        assert environ.get('GRIZZLY_HEALTH_CHECK_RETRIES', None) == '30'
        assert environ.get('LOCUST_WAIT_FOR_WORKERS_REPORT_AFTER_RAMP_UP', None) == '1.25 * WORKER_REPORT_INTERVAL'
        assert environ.get('GRIZZLY_MOUNT_PATH', None) == 'execution-context'

    finally:
        rm_rf(test_context)

        for key in environ:
            if key.startswith('GRIZZLY_'):
                del environ[key]



"""Tests for grizzly_cli.argparse.markdown."""

from __future__ import annotations

import argparse
from typing import TYPE_CHECKING

import pytest
from grizzly_cli.argparse.markdown import MarkdownFormatter, MarkdownHelpAction

if TYPE_CHECKING:
    from _pytest.capture import CaptureFixture
    from pytest_mock import MockerFixture


class TestMarkdownHelpAction:
    def test___init__(self) -> None:
        action = MarkdownHelpAction(['-t', '--test'])

        assert isinstance(action, argparse.Action)
        assert action.option_strings == ['-t', '--test']
        assert action.dest == argparse.SUPPRESS
        assert action.default == argparse.SUPPRESS
        assert action.nargs == 0

    def test___call__(self, mocker: MockerFixture) -> None:
        parser = argparse.ArgumentParser(description='test parser')
        parser.add_argument('--md-help', action=MarkdownHelpAction)

        print_help = mocker.patch.object(parser._actions[-1], 'print_help', autospec=True)

        with pytest.raises(SystemExit) as e:
            parser.parse_args(['--md-help'])
        assert e.type is SystemExit
        assert e.value.code == 0

        assert print_help.call_count == 1
        args, _ = print_help.call_args_list[0]
        assert args[0] is parser

    def test_print_help(self, mocker: MockerFixture) -> None:
        action = MarkdownHelpAction(['-t', '--test'])
        parser = argparse.ArgumentParser(description='test parser')
        parser.add_argument('--md-help', action=MarkdownHelpAction)

        subparsers = parser.add_subparsers(dest='subparser')

        a_parser = subparsers.add_parser('a', description='parser a')
        subparsers.add_parser('b', description='parser b')

        a_subparsers = a_parser.add_subparsers(dest='a_subparsers')
        a_subparsers.add_parser('aa', description='parser aa')

        print_help = mocker.patch('argparse.ArgumentParser.print_help', autospec=True)

        action.print_help(parser)

        assert print_help.call_count == 4
        assert issubclass(parser.formatter_class, MarkdownFormatter)  # type: ignore[arg-type]
        assert parser._subparsers is not None

        _subparsers = getattr(parser, '_subparsers', None)
        assert _subparsers is not None
        for subparsers in _subparsers._group_actions:
            for name, subparser in subparsers.choices.items():
                assert issubclass(subparser.formatter_class, MarkdownFormatter)
                if name == 'a':
                    _subsubparsers = getattr(subparser, '_subparsers', None)
                    assert _subsubparsers is not None
                    for subsubparsers in _subsubparsers._group_actions:
                        for subsubparser in subsubparsers.choices.values():
                            assert issubclass(subsubparser.formatter_class, MarkdownFormatter)

    def test_print_help__format_help_markdown(self, mocker: MockerFixture) -> None:
        action = MarkdownHelpAction(['-t', '--test'])
        parser = argparse.ArgumentParser(description='test parser')
        parser._optionals.title = 'optional arguments'

        formatter = MarkdownFormatter.factory(0)('test-prog')

        _get_formatter = mocker.patch.object(parser, '_get_formatter', side_effect=[formatter])
        add_text = mocker.patch.object(formatter, 'add_text', autospec=True)
        add_usage = mocker.patch.object(formatter, 'add_usage', autospec=True)
        start_section = mocker.patch.object(formatter, 'start_section', autospec=True)
        end_section = mocker.patch.object(formatter, 'end_section', autospec=True)
        add_arguments = mocker.patch.object(formatter, 'add_arguments', autospec=True)

        action.print_help(parser)

        assert _get_formatter.call_count == 1
        assert add_text.call_count == 6
        assert add_usage.call_count == 1
        assert start_section.call_count == 2
        assert start_section.call_args_list[0][0][0] == 'positional arguments'
        assert start_section.call_args_list[1][0][0] == 'optional arguments'
        assert end_section.call_count == 2
        assert add_arguments.call_count == 2


class TestMarkdownFormatter:
    def test___init__(self) -> None:
        formatter = MarkdownFormatter.factory(0)('test')
        assert formatter._root_section is formatter._current_section
        assert formatter._root_section.parent is None
        assert formatter.level == 0
        assert formatter.current_level == 1

    def test__format_usage(self) -> None:
        formatter = MarkdownFormatter.factory(0)('test')
        usage = formatter._format_usage('test', None, None, 'a prefix')
        assert (
            usage
            == """
### Usage

```bash
test
```
"""
        )
        parser = argparse.ArgumentParser(prog='test', description='test parser')
        parser.add_argument('-t', '--test', type=str, required=True, help='test argument')
        parser.add_argument('file', nargs=1, help='file argument')

        core_formatter = parser.formatter_class(prog=parser.prog)

        usage = core_formatter._format_usage(parser.usage, parser._get_positional_actions(), parser._mutually_exclusive_groups, 'a prefix ')
        assert (
            usage
            == """a prefix test file

"""
        )

        usage = formatter._format_usage(parser.usage, parser._get_positional_actions(), parser._mutually_exclusive_groups, 'a prefix ')
        assert (
            usage
            == """
### Usage

```bash
test file
```
"""
        )

    def test_format_help(self) -> None:
        formatter = MarkdownFormatter.factory(0)('test')
        assert formatter.format_help() == ''
        assert formatter._root_section.heading == '# `test`'

    def test_format_text(self) -> None:
        formatter = MarkdownFormatter('test-prog')
        text = """%(prog)s is awesome!
also, here is a sentence. and here is another one!

```bash
hostname -f
```

you cannot belive it, it's another sentence.
"""
        print(formatter._format_text(text))
        assert (
            formatter._format_text(text)
            == """test-prog is awesome!
also, here is a sentence. and here is another one!

```bash
hostname -f
```

you cannot belive it, it's another sentence.
"""
        )

    def test_start_section(self) -> None:
        formatter = MarkdownFormatter.factory(0)('test-prog')
        assert formatter._root_section is formatter._current_section

        formatter.start_section('test-section-01')

        assert formatter._current_section is not formatter._root_section
        assert formatter._current_section.parent is formatter._root_section
        assert formatter._current_section.heading == '## Test-section-01'
        assert len(formatter._current_section.items) == 0
        assert next(iter(formatter._current_section.parent.items)) == (formatter._current_section.format_help, [])

    def test__format_action(self) -> None:
        formatter = MarkdownFormatter.factory(0)('test-prog')
        action = argparse.Action(['-t', '--test'], dest='help', nargs=1, help='test argument')

        assert formatter._format_action(action) == ''

        action.dest = 'test'

        assert formatter._format_action(action) == '| `-t, --test` |  | test argument |\n'

        action.default = 'test-default'
        action.option_strings = ['-t']

        assert formatter._format_action(action) == '| `-t` | `test-default` | test argument |\n'

    def test_current_level(self) -> None:
        formatter = MarkdownFormatter('test-prog')
        formatter.level = 10
        assert formatter.current_level == 11

    class Test_MarkdownSection:
        def test___init__(self) -> None:
            formatter = MarkdownFormatter('test-prog')
            section1 = MarkdownFormatter._MarkdownSection(formatter, None)

            assert section1.formatter is formatter
            assert section1.parent is None
            assert section1.heading is None
            assert len(section1.items) == 0

            section2 = MarkdownFormatter._MarkdownSection(formatter, section1)
            assert section2.formatter is formatter
            assert section2.parent is section1
            assert section2.heading is None
            assert len(section2.items) == 0

            section3 = MarkdownFormatter._MarkdownSection(formatter, section2, 'test heading')
            assert section3.formatter is formatter
            assert section3.parent is section2
            assert section3.heading == 'test heading'
            assert len(section3.items) == 0

        def test_format_help(self, capsys: CaptureFixture) -> None:
            formatter = MarkdownFormatter.factory(0)('test-prog')

            action1 = argparse.Action(['-r', '--root'], dest='root', nargs=2, help='root argument')
            action2 = argparse.Action(['--root-const'], dest='root', nargs=0, default=True)

            formatter.start_section('root section')
            formatter._add_item(formatter._format_action, [action1])
            formatter._add_item(formatter._format_action, [action2])
            formatter.end_section()

            format_help_text = formatter._current_section.format_help()
            assert capsys.readouterr().out == ''
            assert (
                format_help_text
                == """

## Root section

| argument | default | help |
| -------- | ------- | ---- |
| `-r, --root` |  | root argument |
| `--root-const` | `True` |  |


"""
            )
            formatter = MarkdownFormatter('test-prog')
            formatter.level = 1

            action1 = argparse.Action(['-r', '--root'], dest='root', nargs=2, help='root argument')
            action2 = argparse.Action(['--root-const'], dest='root', nargs=0, default=True)

            formatter.start_section('root section')
            formatter._add_item(formatter._format_action, [action1])
            formatter._add_item(formatter._format_action, [action2])
            parent = formatter._current_section.parent
            formatter._current_section.parent = None
            formatter.end_section()
            assert parent is not None
            formatter._current_section = parent
            format_help_text = formatter._current_section.format_help()
            assert capsys.readouterr().out == '\n'  # @TODO: whyyyyyyyyyy?!
            print(format_help_text)
            assert (
                format_help_text
                == """

#### Root section

| argument | default | help |
| -------- | ------- | ---- |
| `-r, --root` |  | root argument |
| `--root-const` | `True` |  |


"""
            )
